from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from guppylang.diagnostic import Error, Help, Note

if TYPE_CHECKING:
    from guppylang.definition.parameter import ParamDef
    from guppylang.tys.ty import Type


@dataclass(frozen=True)
class WrongNumberOfTypeArgsError(Error):
    title: ClassVar[str] = ""  # Custom implementation in `rendered_title`
    expected: int
    actual: int
    type_name: str

    @property
    def rendered_title(self) -> str:
        if self.expected == 0:
            return "Non-parametric type"
        elif self.expected > self.actual:
            return "Missing type arguments"
        else:
            return "Too many type arguments"

    @property
    def rendered_span_label(self) -> str:
        if self.expected == 0:
            return f"Type `{self.type_name}` is not parametric"
        diff = self.expected - self.actual
        msg = "Unexpected " if diff < 0 else "Missing "
        msg += "type arguments " if abs(diff) > 1 else "type argument "
        msg += (
            f"for type `{self.type_name}` (expected {self.expected}, got {self.actual})"
        )
        return msg


@dataclass(frozen=True)
class InvalidTypeArgError(Error):
    title: ClassVar[str] = "Invalid type argument"
    span_label: ClassVar[str] = "Not a valid type argument"


@dataclass(frozen=True)
class IllegalPyTypeArgError(Error):
    title: ClassVar[str] = "Invalid type argument"
    span_label: ClassVar[str] = (
        "Compile-time `py(...)` expression evaluating to `{obj}` is not a valid type "
        "argument"
    )
    obj: object


@dataclass(frozen=True)
class ModuleMemberNotFoundError(Error):
    # TODO: Unify with the definition in expression checker once merged
    title: ClassVar[str] = "Not found in module"
    span_label: ClassVar[str] = "Module `{module_name}` has no member `{member}`"
    module_name: str
    member: str


@dataclass(frozen=True)
class HigherKindedTypeVarError(Error):
    title: ClassVar[str] = "Not parametric"
    span_label: ClassVar[str] = (
        "Type variable `{var_def.name}` doesn't take type arguments"
    )
    var_def: "ParamDef"

    @dataclass(frozen=True)
    class Explain(Note):
        message: ClassVar[str] = "Higher-kinded types are not supported"

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(HigherKindedTypeVarError.Explain(None))


@dataclass(frozen=True)
class FreeTypeVarError(Error):
    title: ClassVar[str] = "Free type variable"
    span_label: ClassVar[str] = "Type variable `{var_def.name}` is unbound"
    var_def: "ParamDef"

    @dataclass(frozen=True)
    class Explain(Note):
        message: ClassVar[str] = (
            "Only struct and function definitions can be generic. Other generic values "
            "or nested types are not supported."
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(FreeTypeVarError.Explain(None))


@dataclass(frozen=True)
class InvalidTypeError(Error):
    title: ClassVar[str] = "Invalid type"
    span_label: ClassVar[str] = "Not a valid type"


@dataclass(frozen=True)
class InvalidCallableTypeError(Error):
    title: ClassVar[str] = "Invalid type"
    span_label: ClassVar[str] = "Invalid function type"

    @dataclass(frozen=True)
    class Explain(Help):
        message: ClassVar[str] = (
            "Function types are specified as follows: "
            "`Callable[[<arguments>], <return type>]`"
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(InvalidCallableTypeError.Explain(None))


@dataclass(frozen=True)
class NonLinearOwnedError(Error):
    title: ClassVar[str] = "Invalid annotation"
    span_label: ClassVar[str] = "Classical type `{ty}` cannot be owned"
    ty: "Type"


@dataclass(frozen=True)
class InvalidFlagError(Error):
    title: ClassVar[str] = "Invalid annotation"
    span_label: ClassVar[str] = "Invalid type annotation"


@dataclass(frozen=True)
class FlagNotAllowedError(Error):
    title: ClassVar[str] = "Invalid annotation"
    span_label: ClassVar[str] = "`@` type annotations are not allowed in this position"



from guppylang.std.builtins import result, py
from tests.util import compile_guppy

# TODO use this once https://github.com/CQCL/guppylang/issues/599 lands
# from guppylang.prelude._internal.checker import TAG_MAX_LEN
# BIG_TAG = "a" * (TAG_MAX_LEN + 1)

@compile_guppy
def foo(y: bool) -> None:
    # each tick or cross is 3 bytes. The cross sends the tag over the limit.
    result("✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅✅❌", y)



import ast
from collections.abc import Iterable, Iterator, Sequence
from contextlib import AbstractContextManager, ExitStack, contextmanager
from typing import Any, Final, TypeGuard, TypeVar

import hugr
import hugr.std.collections.array
import hugr.std.float
import hugr.std.int
import hugr.std.logic
import hugr.std.prelude
from hugr import Wire, ops
from hugr import tys as ht
from hugr import val as hv
from hugr.build import function as hf
from hugr.build.cond_loop import Conditional
from hugr.build.dfg import DP, DfBase
from typing_extensions import assert_never

from guppylang.ast_util import AstNode, AstVisitor, get_type
from guppylang.cfg.builder import tmp_vars
from guppylang.checker.core import Variable, contains_subscript
from guppylang.checker.errors.generic import UnsupportedError
from guppylang.compiler.core import (
    CompilerBase,
    CompilerContext,
    DFContainer,
    GlobalConstId,
)
from guppylang.compiler.hugr_extension import PartialOp
from guppylang.definition.custom import CustomFunctionDef
from guppylang.definition.value import (
    CallReturnWires,
    CompiledCallableDef,
    CompiledValueDef,
)
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.nodes import (
    BarrierExpr,
    DesugaredArrayComp,
    DesugaredGenerator,
    DesugaredListComp,
    FieldAccessAndDrop,
    GenericParamValue,
    GlobalCall,
    GlobalName,
    LocalCall,
    PanicExpr,
    PartialApply,
    PlaceNode,
    ResultExpr,
    SubscriptAccessAndDrop,
    TensorCall,
    TypeApply,
)
from guppylang.std._internal.compiler.arithmetic import convert_ifromusize
from guppylang.std._internal.compiler.array import array_repeat
from guppylang.std._internal.compiler.list import (
    list_new,
)
from guppylang.std._internal.compiler.prelude import build_error, build_panic
from guppylang.tys.arg import Argument
from guppylang.tys.builtin import (
    bool_type,
    get_element_type,
    int_type,
    is_array_type,
    is_bool_type,
)
from guppylang.tys.const import BoundConstVar, ConstValue, ExistentialConstVar
from guppylang.tys.subst import Inst
from guppylang.tys.ty import (
    BoundTypeVar,
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    OpaqueType,
    TupleType,
    Type,
    type_to_row,
)


class ExprCompiler(CompilerBase, AstVisitor[Wire]):
    """A compiler from guppylang expressions to Hugr."""

    dfg: DFContainer

    def compile(self, expr: ast.expr, dfg: DFContainer) -> Wire:
        """Compiles an expression and returns a single wire holding the output value."""
        self.dfg = dfg
        return self.visit(expr)

    def compile_row(self, expr: ast.expr, dfg: DFContainer) -> list[Wire]:
        """Compiles a row expression and returns a list of wires, one for each value in
        the row.

        On Python-level, we treat tuples like rows on top-level. However, nested tuples
        are treated like regular Guppy tuples.
        """
        return [self.compile(e, dfg) for e in expr_to_row(expr)]

    @property
    def builder(self) -> DfBase[ops.DfParentOp]:
        """The current Hugr dataflow graph builder."""
        return self.dfg.builder

    @contextmanager
    def _new_dfcontainer(
        self, inputs: list[PlaceNode], builder: DfBase[DP]
    ) -> Iterator[None]:
        """Context manager to build a graph inside a new `DFContainer`.

        Automatically updates `self.dfg` and makes the inputs available.
        """
        old = self.dfg
        # Check that the input names are unique
        assert len({inp.place.id for inp in inputs}) == len(
            inputs
        ), "Inputs are not unique"
        self.dfg = DFContainer(builder, self.dfg.locals.copy())
        hugr_input = builder.input_node
        for input_node, wire in zip(inputs, hugr_input, strict=True):
            self.dfg[input_node.place] = wire

        yield

        self.dfg = old

    @contextmanager
    def _new_loop(
        self,
        just_inputs_vars: list[PlaceNode],
        loop_vars: list[PlaceNode],
        break_predicate: PlaceNode,
    ) -> Iterator[None]:
        """Context manager to build a graph inside a new `TailLoop` node.

        Automatically adds the `Output` node to the loop body once the context manager
        exits.
        """
        just_inputs = [self.visit(name) for name in just_inputs_vars]
        loop_inputs = [self.visit(name) for name in loop_vars]
        loop = self.builder.add_tail_loop(just_inputs, loop_inputs)
        with self._new_dfcontainer(just_inputs_vars + loop_vars, loop):
            yield
            # Output the branch predicate and the inputs for the next iteration. Note
            # that we have to do fresh calls to `self.visit` here since we're in a new
            # context
            do_break = self.visit(break_predicate)
            loop.set_loop_outputs(do_break, *(self.visit(name) for name in loop_vars))
        # Update the DFG with the outputs from the loop
        for node, wire in zip(loop_vars, loop, strict=True):
            self.dfg[node.place] = wire

    @contextmanager
    def _new_case(
        self,
        inputs: list[PlaceNode],
        outputs: list[PlaceNode],
        conditional: Conditional,
        case_id: int,
    ) -> Iterator[None]:
        """Context manager to build a graph inside a new `Case` node.

        Automatically adds the `Output` node once the context manager exits.
        """
        # TODO: `Case` is `_DfgBase`, but not `Dfg`?
        case = conditional.add_case(case_id)
        with self._new_dfcontainer(inputs, case):
            yield
            case.set_outputs(*(self.visit(name) for name in outputs))

    def _if_else(
        self,
        cond: ast.expr,
        inputs: list[PlaceNode],
        outputs: list[PlaceNode],
        only_true_inputs: list[PlaceNode] | None = None,
        only_false_inputs: list[PlaceNode] | None = None,
    ) -> tuple[AbstractContextManager[None], AbstractContextManager[None]]:
        """Builds a `Conditional`, returning context managers to build the `True` and
        `False` branch.
        """
        conditional = self.builder.add_conditional(
            self.visit(cond), *(self.visit(inp) for inp in inputs)
        )
        only_true_inputs_ = only_true_inputs or []
        only_false_inputs_ = only_false_inputs or []

        @contextmanager
        def true_case() -> Iterator[None]:
            with self._new_case(only_true_inputs_ + inputs, outputs, conditional, 1):
                yield

        @contextmanager
        def false_case() -> Iterator[None]:
            with self._new_case(only_false_inputs_ + inputs, outputs, conditional, 0):
                yield
            # Update the DFG with the outputs from the Conditional node
            for node, wire in zip(outputs, conditional, strict=True):
                self.dfg[node.place] = wire

        return true_case(), false_case()

    @contextmanager
    def _if_true(self, cond: ast.expr, inputs: list[PlaceNode]) -> Iterator[None]:
        """Context manager to build a graph inside the `true` case of a `Conditional`

        In the `false` case, the inputs are outputted as is.
        """
        true_case, false_case = self._if_else(cond, inputs, inputs)
        with false_case:
            # If the condition is false, output the inputs as is
            pass
        with true_case:
            # If the condition is true, we enter the `with` block
            yield

    def visit_Constant(self, node: ast.Constant) -> Wire:
        if value := python_value_to_hugr(node.value, get_type(node)):
            return self.builder.load(value)
        raise InternalGuppyError("Unsupported constant expression in compiler")

    def visit_PlaceNode(self, node: PlaceNode) -> Wire:
        if subscript := contains_subscript(node.place):
            if subscript.item not in self.dfg:
                self.dfg[subscript.item] = self.visit(subscript.item_expr)
            self.dfg[subscript] = self.visit(subscript.getitem_call)
        return self.dfg[node.place]

    def visit_GlobalName(self, node: GlobalName) -> Wire:
        defn = self.ctx.build_compiled_def(node.def_id)
        assert isinstance(defn, CompiledValueDef)
        if isinstance(defn, CompiledCallableDef) and defn.ty.parametrized:
            # TODO: This should be caught during checking
            err = UnsupportedError(
                node, "Polymorphic functions as dynamic higher-order values"
            )
            raise GuppyError(err)
        return defn.load(self.dfg, self.ctx, node)

    def visit_GenericParamValue(self, node: GenericParamValue) -> Wire:
        match node.param.ty:
            case NumericType(NumericType.Kind.Nat):
                arg = node.param.to_bound().to_hugr()
                load_nat = hugr.std.PRELUDE.get_op("load_nat").instantiate(
                    [arg], ht.FunctionType([], [ht.USize()])
                )
                usize = self.builder.add_op(load_nat)
                return self.builder.add_op(convert_ifromusize(), usize)
            case _:
                raise NotImplementedError

    def visit_Name(self, node: ast.Name) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_Tuple(self, node: ast.Tuple) -> Wire:
        elems = [self.visit(e) for e in node.elts]
        types = [get_type(e) for e in node.elts]
        return self._pack_tuple(elems, types)

    def visit_List(self, node: ast.List) -> Wire:
        # Note that this is a list literal (i.e. `[e1, e2, ...]`), not a comprehension
        inputs = [self.visit(e) for e in node.elts]
        list_ty = get_type(node)
        elem_ty = get_element_type(list_ty)
        return list_new(self.builder, elem_ty.to_hugr(), inputs)

    def _unpack_tuple(self, wire: Wire, types: Sequence[Type]) -> Sequence[Wire]:
        """Add a tuple unpack operation to the graph"""
        types = [t.to_hugr() for t in types]
        return list(self.builder.add_op(ops.UnpackTuple(types), wire))

    def _pack_tuple(self, wires: Sequence[Wire], types: Sequence[Type]) -> Wire:
        """Add a tuple pack operation to the graph"""
        types = [t.to_hugr() for t in types]
        return self.builder.add_op(ops.MakeTuple(types), *wires)

    def _pack_returns(self, returns: Sequence[Wire], return_ty: Type) -> Wire:
        """Groups function return values into a tuple"""
        if isinstance(return_ty, TupleType | NoneType) and not return_ty.preserve:
            types = type_to_row(return_ty)
            assert len(returns) == len(types)
            return self._pack_tuple(returns, types)
        assert (
            len(returns) == 1
        ), f"Expected a single return value. Got {returns}. return type {return_ty}"
        return returns[0]

    def _update_inout_ports(
        self,
        args: list[ast.expr],
        inout_ports: Iterator[Wire],
        func_ty: FunctionType,
    ) -> None:
        """Helper method that updates the ports for borrowed arguments after a call."""
        for inp, arg in zip(func_ty.inputs, args, strict=True):
            if InputFlags.Inout in inp.flags:
                # Linearity checker ensures that borrowed arguments that are not places
                # can be safely dropped after the call returns
                if not isinstance(arg, PlaceNode):
                    next(inout_ports)
                    continue
                self.dfg[arg.place] = next(inout_ports)
                # Places involving subscripts need to generate code for the appropriate
                # `__setitem__` call. Nested subscripts are handled automatically since
                # `arg.place.parent` occurs as an arg of this call, so will also
                # be recursively reassigned.
                if subscript := contains_subscript(arg.place):
                    assert subscript.setitem_call is not None
                    # Need to assign __setitem__ value before compiling call.
                    # Note that the assignment to `self.dfg[arg.place]` also updated
                    # `self.dfg[subscript]` so that it now contains the value we want
                    # to write back into the subscript.
                    self.dfg[subscript.setitem_call.value_var] = self.dfg[subscript]
                    self.visit(subscript.setitem_call.call)
        assert next(inout_ports, None) is None, "Too many inout return ports"

    def visit_LocalCall(self, node: LocalCall) -> Wire:
        func = self.visit(node.func)
        func_ty = get_type(node.func)
        assert isinstance(func_ty, FunctionType)
        num_returns = len(type_to_row(func_ty.output))

        args = [self.visit(arg) for arg in node.args]
        call = self.builder.add_op(ops.CallIndirect(func_ty.to_hugr()), func, *args)
        regular_returns = list(call[:num_returns])
        inout_returns = call[num_returns:]
        self._update_inout_ports(node.args, inout_returns, func_ty)
        return self._pack_returns(regular_returns, func_ty.output)

    def visit_TensorCall(self, node: TensorCall) -> Wire:
        functions: Wire = self.visit(node.func)
        function_types = get_type(node.func)
        assert isinstance(function_types, TupleType)

        rets: list[Wire] = []
        remaining_args = node.args
        for func, func_ty in zip(
            self._unpack_tuple(functions, function_types.element_types),
            function_types.element_types,
            strict=True,
        ):
            outs, remaining_args = self._compile_tensor_with_leftovers(
                func, func_ty, remaining_args
            )
            rets.extend(outs)
        assert (
            remaining_args == []
        ), "Not all function arguments were consumed after a tensor call"
        return self._pack_returns(rets, node.tensor_ty.output)

    def _compile_tensor_with_leftovers(
        self, func: Wire, func_ty: Type, args: list[ast.expr]
    ) -> tuple[
        list[Wire],  # Compiled outputs
        list[ast.expr],  # Leftover args
    ]:
        """Compiles a function call, consuming as many arguments as needed, and
        returning the unused ones.
        """
        if isinstance(func_ty, TupleType):
            remaining_args = args
            all_outs = []
            for elem, ty in zip(
                self._unpack_tuple(func, func_ty.element_types),
                func_ty.element_types,
                strict=True,
            ):
                outs, remaining_args = self._compile_tensor_with_leftovers(
                    elem, ty, remaining_args
                )
                all_outs.extend(outs)
            return all_outs, remaining_args

        elif isinstance(func_ty, FunctionType):
            input_len = len(func_ty.inputs)
            num_returns = len(type_to_row(func_ty.output))
            consumed_args, other_args = args[0:input_len], args[input_len:]
            consumed_wires = [self.visit(arg) for arg in consumed_args]
            call = self.builder.add_op(
                ops.CallIndirect(func_ty.to_hugr()), func, *consumed_wires
            )
            regular_returns: list[Wire] = list(call[:num_returns])
            inout_returns = call[num_returns:]
            self._update_inout_ports(consumed_args, inout_returns, func_ty)
            return regular_returns, other_args
        else:
            raise InternalGuppyError("Tensor element wasn't function or tuple")

    def visit_GlobalCall(self, node: GlobalCall) -> Wire:
        func = self.ctx.build_compiled_def(node.def_id)
        assert isinstance(func, CompiledCallableDef)

        args = [self.visit(arg) for arg in node.args]
        rets = func.compile_call(args, list(node.type_args), self.dfg, self.ctx, node)
        if isinstance(func, CustomFunctionDef) and not func.has_signature:
            func_ty = FunctionType(
                [FuncInput(get_type(arg), InputFlags.NoFlags) for arg in node.args],
                get_type(node),
            )
        else:
            func_ty = func.ty.instantiate(node.type_args)
        self._update_inout_ports(node.args, iter(rets.inout_returns), func_ty)
        return self._pack_returns(rets.regular_returns, func_ty.output)

    def visit_Call(self, node: ast.Call) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_PartialApply(self, node: PartialApply) -> Wire:
        func_ty = get_type(node.func)
        assert isinstance(func_ty, FunctionType)
        op = PartialOp.from_closure(
            func_ty.to_hugr(), [get_type(arg).to_hugr() for arg in node.args]
        )
        return self.builder.add_op(
            op, self.visit(node.func), *(self.visit(arg) for arg in node.args)
        )

    def visit_TypeApply(self, node: TypeApply) -> Wire:
        # For now, we can only TypeApply global FunctionDefs/Decls.
        if not isinstance(node.value, GlobalName):
            raise InternalGuppyError("Dynamic TypeApply not supported yet!")
        defn = self.ctx.build_compiled_def(node.value.def_id)
        assert isinstance(defn, CompiledCallableDef)

        # We have to be very careful here: If we instantiate `foo: forall T. T -> T`
        # with a tuple type `tuple[A, B]`, we get the type `tuple[A, B] -> tuple[A, B]`.
        # Normally, this would be represented in Hugr as a function with two output
        # ports types A and B. However, when TypeApplying `foo`, we actually get a
        # function with a single output port typed `tuple[A, B]`.
        # TODO: We would need to do manual monomorphisation in that case to obtain a
        #  function that returns two ports as expected
        if instantiation_needs_unpacking(defn.ty, node.inst):
            err = UnsupportedError(
                node, "Generic function instantiations returning rows"
            )
            raise GuppyError(err)

        return defn.load_with_args(node.inst, self.dfg, self.ctx, node)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> Wire:
        # The only case that is not desugared by the type checker is the `not` operation
        # since it is not implemented via a dunder method
        if isinstance(node.op, ast.Not):
            arg = self.visit(node.operand)
            return self.builder.add_op(hugr.std.logic.Not, arg)

        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_FieldAccessAndDrop(self, node: FieldAccessAndDrop) -> Wire:
        struct_port = self.visit(node.value)
        field_idx = node.struct_ty.fields.index(node.field)
        return self._unpack_tuple(struct_port, [f.ty for f in node.struct_ty.fields])[
            field_idx
        ]

    def visit_SubscriptAccessAndDrop(self, node: SubscriptAccessAndDrop) -> Wire:
        self.dfg[node.item] = self.visit(node.item_expr)
        return self.visit(node.getitem_expr)

    def visit_ResultExpr(self, node: ResultExpr) -> Wire:
        extra_args = []
        if isinstance(node.base_ty, NumericType):
            match node.base_ty.kind:
                case NumericType.Kind.Nat:
                    base_name = "uint"
                    extra_args = [ht.BoundedNatArg(n=NumericType.INT_WIDTH)]
                case NumericType.Kind.Int:
                    base_name = "int"
                    extra_args = [ht.BoundedNatArg(n=NumericType.INT_WIDTH)]
                case NumericType.Kind.Float:
                    base_name = "f64"
                case kind:
                    assert_never(kind)
        else:
            # The only other valid base type is bool
            assert is_bool_type(node.base_ty)
            base_name = "bool"
        if node.array_len is not None:
            op_name = f"result_array_{base_name}"
            match node.array_len:
                case ConstValue(value=value):
                    assert isinstance(value, int)
                    extra_args = [ht.BoundedNatArg(n=value), *extra_args]
                case BoundConstVar():
                    # TODO: We need to handle this once we allow function definitions
                    #  that are generic over array lengths
                    raise NotImplementedError
                case ExistentialConstVar() as var:
                    raise InternalGuppyError(
                        f"Unsolved existential variable during Hugr lowering: {var}"
                    )
                case c:
                    assert_never(c)
        else:
            op_name = f"result_{base_name}"
        op = tket2_result_op(
            op_name=op_name,
            typ=get_type(node.value).to_hugr(),
            tag=node.tag,
            extra_args=extra_args,
        )
        self.builder.add_op(op, self.visit(node.value))
        return self._pack_returns([], NoneType())

    def visit_PanicExpr(self, node: PanicExpr) -> Wire:
        err = build_error(self.builder, 1, node.msg)
        in_tys = [get_type(e).to_hugr() for e in node.values]
        out_tys = [ty.to_hugr() for ty in type_to_row(get_type(node))]
        outs = build_panic(
            self.builder,
            in_tys,
            out_tys,
            err,
            *(self.visit(e) for e in node.values),
        ).outputs()
        return self._pack_returns(list(outs), get_type(node))

    def visit_BarrierExpr(self, node: BarrierExpr) -> Wire:
        hugr_tys = [get_type(e).to_hugr() for e in node.args]
        op = hugr.std.prelude.PRELUDE_EXTENSION.get_op("Barrier").instantiate(
            [ht.SequenceArg([ht.TypeTypeArg(ty) for ty in hugr_tys])],
            ht.FunctionType.endo(hugr_tys),
        )

        barrier_n = self.builder.add_op(op, *(self.visit(e) for e in node.args))

        self._update_inout_ports(node.args, iter(barrier_n), node.func_ty)
        return self._pack_returns([], NoneType())

    def visit_DesugaredListComp(self, node: DesugaredListComp) -> Wire:
        # Make up a name for the list under construction and bind it to an empty list
        list_ty = get_type(node)
        assert isinstance(list_ty, OpaqueType)
        elem_ty = get_element_type(list_ty)
        list_place = Variable(next(tmp_vars), list_ty, node)
        self.dfg[list_place] = list_new(self.builder, elem_ty.to_hugr(), [])
        with self._build_generators(node.generators, [list_place]):
            elt_port = self.visit(node.elt)
            list_port = self.dfg[list_place]
            [], [self.dfg[list_place]] = self._build_method_call(
                list_ty, "append", node, [list_port, elt_port], list_ty.args
            )
        return self.dfg[list_place]

    def visit_DesugaredArrayComp(self, node: DesugaredArrayComp) -> Wire:
        # Allocate an uninitialised array of the desired size and a counter variable
        array_ty = get_type(node)
        assert isinstance(array_ty, OpaqueType)
        array_var = Variable(next(tmp_vars), array_ty, node)
        count_var = Variable(next(tmp_vars), int_type(), node)
        # See https://github.com/CQCL/guppylang/issues/629
        hugr_elt_ty = ht.Option(node.elt_ty.to_hugr())
        # Initialise array with `None`s
        make_none = array_comprehension_init_func(self.ctx)
        make_none = self.builder.load_function(
            make_none,
            instantiation=ht.FunctionType([], [hugr_elt_ty]),
            type_args=[ht.TypeTypeArg(node.elt_ty.to_hugr())],
        )
        self.dfg[array_var] = self.builder.add_op(
            array_repeat(hugr_elt_ty, node.length.to_arg().to_hugr()), make_none
        )
        self.dfg[count_var] = self.builder.load(
            hugr.std.int.IntVal(0, width=NumericType.INT_WIDTH)
        )
        with self._build_generators([node.generator], [array_var, count_var]):
            elt = self.visit(node.elt)
            array, count = self.dfg[array_var], self.dfg[count_var]
            [], [self.dfg[array_var]] = self._build_method_call(
                array_ty, "__setitem__", node, [array, count, elt], array_ty.args
            )
            # Update `count += 1`
            one = self.builder.load(hugr.std.int.IntVal(1, width=NumericType.INT_WIDTH))
            [self.dfg[count_var]], [] = self._build_method_call(
                int_type(), "__add__", node, [count, one]
            )
        return self.dfg[array_var]

    def _build_method_call(
        self,
        ty: Type,
        method: str,
        node: AstNode,
        args: list[Wire],
        type_args: Sequence[Argument] | None = None,
    ) -> CallReturnWires:
        func = self.ctx.get_instance_func(ty, method)
        assert func is not None
        return func.compile_call(args, type_args or [], self.dfg, self.ctx, node)

    @contextmanager
    def _build_generators(
        self, gens: list[DesugaredGenerator], loop_vars: list[Variable]
    ) -> Iterator[None]:
        """Context manager to build and enter the `TailLoop`s for a list of generators.

        The provided `loop_vars` will be threaded through and will be available inside
        the loops.
        """
        from guppylang.compiler.stmt_compiler import StmtCompiler

        compiler = StmtCompiler(self.ctx)
        with ExitStack() as stack:
            for gen in gens:
                # Build the generator
                compiler.compile_stmts([gen.iter_assign], self.dfg)
                assert isinstance(gen.iter, PlaceNode)
                iter_ty = get_type(gen.iter)
                inputs = [PlaceNode(place=var) for var in loop_vars]
                inputs += [
                    PlaceNode(place=place) for place in gen.borrowed_outer_places
                ]
                # Enter a new tail loop. Note that the iterator is a `just_input`, so
                # will not be outputted by the loop
                break_pred = PlaceNode(Variable(next(tmp_vars), bool_type(), gen.iter))
                stack.enter_context(self._new_loop([gen.iter], inputs, break_pred))
                # Enter a conditional checking if we have a next element
                next_ty = TupleType([get_type(gen.target), iter_ty])
                next_var = PlaceNode(Variable(next(tmp_vars), next_ty, gen.iter))
                hasnext_case, stop_case = self._if_else(
                    gen.next_call,
                    inputs,
                    only_true_inputs=[next_var],
                    outputs=[break_pred, *inputs],
                )
                # In the "no" case, we set the break predicate to true
                break_pred_hugr_ty = ht.Either([iter_ty.to_hugr()], [])
                with stop_case:
                    self.dfg[break_pred.place] = self.dfg.builder.add_op(
                        ops.Tag(1, break_pred_hugr_ty)
                    )
                # Otherwise, we continue, set the break predicate to false, and insert
                # the iterator for the next loop iteration
                stack.enter_context(hasnext_case)
                next_wire = self.dfg[next_var.place]
                elt, it = self.dfg.builder.add_op(ops.UnpackTuple(), next_wire)
                compiler.dfg = self.dfg
                compiler._assign(gen.target, elt)
                self.dfg[break_pred.place] = self.dfg.builder.add_op(
                    ops.Tag(0, break_pred_hugr_ty), it
                )
                # Enter nested conditionals for each if guard on the generator
                for if_expr in gen.ifs:
                    stack.enter_context(self._if_true(if_expr, [break_pred, *inputs]))
            # Yield control to the caller to build inside the loop
            yield

    def visit_BinOp(self, node: ast.BinOp) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_Compare(self, node: ast.Compare) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")


def expr_to_row(expr: ast.expr) -> list[ast.expr]:
    """Turns an expression into a row expressions by unpacking top-level tuples."""
    return expr.elts if isinstance(expr, ast.Tuple) else [expr]


def instantiation_needs_unpacking(func_ty: FunctionType, inst: Inst) -> bool:
    """Checks if instantiating a polymorphic makes it return a row."""
    if isinstance(func_ty.output, BoundTypeVar):
        return_ty = inst[func_ty.output.idx]
        return isinstance(return_ty, TupleType | NoneType)
    return False


def python_value_to_hugr(v: Any, exp_ty: Type) -> hv.Value | None:
    """Turns a Python value into a Hugr value.

    Returns None if the Python value cannot be represented in Guppy.
    """
    match v:
        case bool():
            return hv.bool_value(v)
        case str():
            return hugr.std.prelude.StringVal(v)
        case int():
            return hugr.std.int.IntVal(v, width=NumericType.INT_WIDTH)
        case float():
            return hugr.std.float.FloatVal(v)
        case tuple(elts):
            assert isinstance(exp_ty, TupleType)
            vs = [
                python_value_to_hugr(elt, ty)
                for elt, ty in zip(elts, exp_ty.element_types, strict=True)
            ]
            if doesnt_contain_none(vs):
                return hv.Tuple(*vs)
        case list(elts):
            assert is_array_type(exp_ty)
            elem_ty = get_element_type(exp_ty)
            vs = [python_value_to_hugr(elt, elem_ty) for elt in elts]
            if doesnt_contain_none(vs):
                opt_ty = ht.Option(elem_ty.to_hugr())
                opt_vs: list[hv.Value] = [hv.Some(v) for v in vs]
                return hugr.std.collections.array.ArrayVal(opt_vs, opt_ty)
        case _:
            return None
    return None


def tket2_result_op(
    op_name: str,
    typ: ht.Type,
    tag: str,
    extra_args: Iterable[ht.TypeArg],
) -> ops.DataflowOp:
    """Creates a dummy operation for constructing a list."""
    args = [
        ht.StringArg(tag),
        *extra_args,
    ]
    sig = ht.FunctionType(
        input=[typ],
        output=[],
    )
    return ops.Custom(
        extension="tket2.result",
        op_name=op_name,
        args=args,
        signature=sig,
    )


ARRAY_COMPREHENSION_INIT: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__comprehension.init"
)


def array_comprehension_init_func(ctx: CompilerContext) -> hf.Function:
    """Returns the Hugr function that is used to initialise arrays elements before a
    comprehension.

    Just returns the `None` variant of the optional element type.

    See https://github.com/CQCL/guppylang/issues/629
    """
    v = ht.Variable(0, ht.TypeBound(ht.TypeBound.Any))
    sig = ht.PolyFuncType(
        params=[ht.TypeTypeParam(ht.TypeBound.Any)],
        body=ht.FunctionType([], [ht.Option(v)]),
    )
    func, already_defined = ctx.declare_global_func(ARRAY_COMPREHENSION_INIT, sig)
    if not already_defined:
        func.set_outputs(func.add_op(ops.Tag(0, ht.Option(v))))
    return func


T = TypeVar("T")


def doesnt_contain_none(xs: list[T | None]) -> TypeGuard[list[T]]:
    """Checks if a list contains `None`."""
    return all(x is not None for x in xs)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

mod = GuppyModule("test")

@guppy(mod)
def x() -> None:
    pass

@guppy(mod)
def bad(b: bool) -> int:
    if b:
        x = 4
    return x

mod.compile()



import ast
import inspect
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

import hugr.build.function as hf
import hugr.tys as ht
from hugr import Wire
from hugr.build.dfg import DefinitionBuilder, OpVar
from hugr.hugr.node_port import ToNode
from hugr.package import FuncDefnPointer

from guppylang.ast_util import AstNode, annotate_location, with_loc, with_type
from guppylang.checker.cfg_checker import CheckedCFG
from guppylang.checker.core import Context, Globals, Place, PyScope
from guppylang.checker.errors.generic import ExpectedError
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import (
    check_global_func_def,
    check_signature,
    parse_function_with_docstring,
)
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.compiler.func_compiler import compile_global_func_def
from guppylang.definition.common import (
    CheckableDef,
    CompilableDef,
    ParsableDef,
    UnknownSourceError,
)
from guppylang.definition.value import CallableDef, CallReturnWires, CompiledCallableDef
from guppylang.error import GuppyError
from guppylang.ipython_inspect import find_ipython_def, is_running_ipython
from guppylang.nodes import GlobalCall
from guppylang.span import SourceMap
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import FunctionType, Type, type_to_row

PyFunc = Callable[..., Any]


@dataclass(frozen=True)
class RawFunctionDef(ParsableDef):
    """A raw function definition provided by the user.

    The raw definition stores exactly what the user has written (i.e. the AST), without
    any additional checking or parsing. Furthermore, we store the values of the Python
    variables in scope at the point of definition.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        python_func: The Python function to be defined.
        python_scope: The Python scope where the function was defined.
    """

    python_func: PyFunc
    python_scope: PyScope = field(repr=False)

    description: str = field(default="function", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ParsedFunctionDef":
        """Parses and checks the user-provided signature of the function."""
        func_ast, docstring = parse_py_func(self.python_func, sources)
        ty = check_signature(func_ast, globals.with_python_scope(self.python_scope))
        return ParsedFunctionDef(
            self.id, self.name, func_ast, ty, self.python_scope, docstring
        )

    def compile(self) -> FuncDefnPointer:
        from guppylang.decorator import guppy

        return guppy.compile_function(self)


@dataclass(frozen=True)
class ParsedFunctionDef(CheckableDef, CallableDef):
    """A function definition with parsed and checked signature.

    In particular, this means that we have determined a type for the function and are
    ready to check the function body.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        ty: The type of the function.
        python_scope: The Python scope where the function was defined.
        docstring: The docstring of the function.
    """

    python_scope: PyScope = field(repr=False)
    defined_at: ast.FunctionDef
    ty: FunctionType
    docstring: str | None

    description: str = field(default="function", init=False)

    def check(self, globals: Globals) -> "CheckedFunctionDef":
        """Type checks the body of the function."""
        # Add python variable scope to the globals
        globals = globals.with_python_scope(self.python_scope)
        cfg = check_global_func_def(self.defined_at, self.ty, globals)
        return CheckedFunctionDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.python_scope,
            self.docstring,
            cfg,
        )

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call."""
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return with_type(ty, node), ty


@dataclass(frozen=True)
class CheckedFunctionDef(ParsedFunctionDef, CompilableDef):
    """Type checked version of a user-defined function that is ready to be compiled.

    In particular, this means that we have a constructed and type checked a control-flow
    graph for the function body.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        ty: The type of the function.
        python_scope: The Python scope where the function was defined.
        docstring: The docstring of the function.
        cfg: The type- and linearity-checked CFG for the function body.
    """

    cfg: CheckedCFG[Place]

    def compile_outer(self, module: DefinitionBuilder[OpVar]) -> "CompiledFunctionDef":
        """Adds a Hugr `FuncDefn` node for this function to the Hugr.

        Note that we don't compile the function body at this point since we don't have
        access to the other compiled functions yet. The body is compiled later in
        `CompiledFunctionDef.compile_inner()`.
        """
        func_type = self.ty.to_hugr_poly()
        func_def = module.define_function(
            self.name, func_type.body.input, func_type.body.output, func_type.params
        )
        return CompiledFunctionDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.python_scope,
            self.docstring,
            self.cfg,
            func_def,
        )


@dataclass(frozen=True)
class CompiledFunctionDef(CheckedFunctionDef, CompiledCallableDef):
    """A function definition with a corresponding Hugr node.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        ty: The type of the function.
        python_scope: The Python scope where the function was defined.
        docstring: The docstring of the function.
        cfg: The type- and linearity-checked CFG for the function body.
        func_def: The Hugr function definition.
    """

    func_def: hf.Function

    def load_with_args(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the function as a value into a local Hugr dataflow graph."""
        return load_with_args(type_args, dfg, self.ty, self.func_def)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        return compile_call(args, type_args, dfg, self.ty, self.func_def)

    def compile_inner(self, globals: CompilerContext) -> None:
        """Compiles the body of the function."""
        compile_global_func_def(self, self.func_def, globals)


def load_with_args(
    type_args: Inst,
    dfg: DFContainer,
    ty: FunctionType,
    func: ToNode,
) -> Wire:
    """Loads the function as a value into a local Hugr dataflow graph."""
    func_ty: ht.FunctionType = ty.instantiate(type_args).to_hugr()
    type_args: list[ht.TypeArg] = [arg.to_hugr() for arg in type_args]
    return dfg.builder.load_function(func, func_ty, type_args)


def compile_call(
    args: list[Wire],
    type_args: Inst,
    dfg: DFContainer,
    ty: FunctionType,
    func: ToNode,
) -> CallReturnWires:
    """Compiles a call to the function."""
    func_ty: ht.FunctionType = ty.instantiate(type_args).to_hugr()
    type_args: list[ht.TypeArg] = [arg.to_hugr() for arg in type_args]
    num_returns = len(type_to_row(ty.output))
    call = dfg.builder.call(func, *args, instantiation=func_ty, type_args=type_args)
    return CallReturnWires(
        regular_returns=list(call[:num_returns]),
        inout_returns=list(call[num_returns:]),
    )


def parse_py_func(f: PyFunc, sources: SourceMap) -> tuple[ast.FunctionDef, str | None]:
    source_lines, line_offset = inspect.getsourcelines(f)
    source, func_ast, line_offset = parse_source(source_lines, line_offset)
    # In Jupyter notebooks, we shouldn't use `inspect.getsourcefile(f)` since it would
    # only give us a dummy temporary file
    file: str | None
    if is_running_ipython():
        file = "<In [?]>"
        if isinstance(func_ast, ast.FunctionDef):
            defn = find_ipython_def(func_ast.name)
            if defn is not None:
                file = f"<{defn.cell_name}>"
                sources.add_file(file, defn.cell_source)
            else:
                # If we couldn't find the defining cell, just use the source code we
                # got from inspect. Line numbers will be wrong, but that's the best we
                # can do.
                sources.add_file(file, source)
                line_offset = 1
    else:
        file = inspect.getsourcefile(f)
        if file is None:
            raise GuppyError(UnknownSourceError(None, f))
        sources.add_file(file)
    annotate_location(func_ast, source, file, line_offset)
    if not isinstance(func_ast, ast.FunctionDef):
        raise GuppyError(ExpectedError(func_ast, "a function definition"))
    return parse_function_with_docstring(func_ast)


def parse_source(source_lines: list[str], line_offset: int) -> tuple[str, ast.AST, int]:
    """Parses a list of source lines into an AST object.

    Also takes care of correctly parsing source that is indented.

    Returns the full source, the parsed AST node, and a potentially updated line number
    offset.
    """
    source = "".join(source_lines)  # Lines already have trailing \n's
    if source_lines[0][0].isspace():
        # This means the function is indented, so we cannot parse it straight away.
        # Running `textwrap.dedent` would mess up the column number in spans. Instead,
        # we'll just wrap the source into a dummy class definition so the indent becomes
        # valid
        cls_node = ast.parse("class _:\n" + source).body[0]
        assert isinstance(cls_node, ast.ClassDef)
        node = cls_node.body[0]
        line_offset -= 1
    else:
        node = ast.parse(source).body[0]
    return source, node, line_offset



"""A hugr extension with guppy-specific operations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import hugr.ext as he
import hugr.tys as ht
from hugr import ops

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

EXTENSION: he.Extension = he.Extension("guppylang", he.Version(0, 1, 0))


PARTIAL_OP_DEF: he.OpDef = EXTENSION.add_op_def(
    he.OpDef(
        "partial",
        signature=he.OpDefSig(
            poly_func=ht.PolyFuncType(
                params=[
                    # Captured input types
                    ht.ListParam(ht.TypeTypeParam(ht.TypeBound.Any)),
                    # Non-captured input types
                    ht.ListParam(ht.TypeTypeParam(ht.TypeBound.Any)),
                    # Output types
                    ht.ListParam(ht.TypeTypeParam(ht.TypeBound.Any)),
                ],
                body=ht.FunctionType(
                    input=[
                        ht.FunctionType(
                            input=[
                                ht.RowVariable(0, ht.TypeBound.Any),
                                ht.RowVariable(1, ht.TypeBound.Any),
                            ],
                            output=[ht.RowVariable(2, ht.TypeBound.Any)],
                        ),
                        ht.RowVariable(0, ht.TypeBound.Any),
                    ],
                    output=[
                        ht.FunctionType(
                            input=[ht.RowVariable(1, ht.TypeBound.Any)],
                            output=[ht.RowVariable(2, ht.TypeBound.Any)],
                        ),
                    ],
                ),
            )
        ),
        description="A partial application of a function."
        " Given arguments [*a],[*b],[*c], represents an operation with type"
        " `(*c, *a -> *b), *c -> (*a -> *b)`",
    )
)


@dataclass
class PartialOp(ops.AsExtOp):
    """An operation that partially evaluates a function.

    args:
        captured_inputs: A list of input types `c_0, ..., c_k` to partially apply.
        other_inputs: A list of input types `a_0, ..., a_n` not partially applied.
        outputs: The output types `b_0, ..., b_m` of the partially applied function.

    returns:
        An operation with type
            ` (c_0, ..., c_k, a_0, ..., a_n -> b_0, ..., b_m ), c_0, ..., c_k`
            `-> (a_0, ..., a_n -> b_0, ..., b_m)`
    """

    captured_inputs: list[ht.Type]
    other_inputs: list[ht.Type]
    outputs: list[ht.Type]

    @classmethod
    def from_closure(
        cls, closure_ty: ht.FunctionType, captured_tys: Sequence[ht.Type]
    ) -> PartialOp:
        """An operation that partially evaluates a function.

        args:
            closure_ty: A function `(c_0, ..., c_k, a_0, ..., a_n) -> b_0, ..., b_m`
            captured_tys: A list `c_0, ..., c_k` of types captured by the function

        returns:
            An operation with type
                ` (c_0, ..., c_k, a_0, ..., a_n -> b_0, ..., b_m ), c_0, ..., c_k`
                `-> (a_0, ..., a_n -> b_0, ..., b_m)`
        """
        assert len(closure_ty.input) >= len(captured_tys)
        assert captured_tys == closure_ty.input[: len(captured_tys)]

        other_inputs = closure_ty.input[len(captured_tys) :]
        return cls(
            captured_inputs=list(captured_tys),
            other_inputs=list(other_inputs),
            outputs=list(closure_ty.output),
        )

    def op_def(self) -> he.OpDef:
        return PARTIAL_OP_DEF

    def type_args(self) -> list[ht.TypeArg]:
        captured_args: list[ht.TypeArg] = [
            ht.TypeTypeArg(ty) for ty in self.captured_inputs
        ]
        other_args: list[ht.TypeArg] = [ht.TypeTypeArg(ty) for ty in self.other_inputs]
        output_args: list[ht.TypeArg] = [ht.TypeTypeArg(ty) for ty in self.outputs]
        return [
            ht.SequenceArg(captured_args),
            ht.SequenceArg(other_args),
            ht.SequenceArg(output_args),
        ]

    def cached_signature(self) -> ht.FunctionType | None:
        closure_ty = ht.FunctionType(
            [*self.captured_inputs, *self.other_inputs],
            self.outputs,
        )
        partial_fn_ty = ht.FunctionType(self.other_inputs, closure_ty.output)
        return ht.FunctionType([closure_ty, *self.captured_inputs], [partial_fn_ty])

    @classmethod
    def from_ext(cls, custom: ops.ExtOp) -> PartialOp:
        match custom:
            case ops.ExtOp(
                _op_def=op_def, args=[captured_args, other_args, output_args]
            ):
                if op_def.qualified_name() == PARTIAL_OP_DEF.qualified_name():
                    return cls(
                        captured_inputs=[*_arg_seq_to_types(captured_args)],
                        other_inputs=[*_arg_seq_to_types(other_args)],
                        outputs=[*_arg_seq_to_types(output_args)],
                    )
        msg = f"Invalid custom op: {custom}"
        raise ops.AsExtOp.InvalidExtOp(msg)

    @property
    def num_out(self) -> int:
        return 1


UNSUPPORTED_OP_DEF: he.OpDef = EXTENSION.add_op_def(
    he.OpDef(
        "unsupported",
        signature=he.OpDefSig(
            poly_func=ht.PolyFuncType(
                params=[
                    # Name of the operation
                    ht.StringParam(),
                    # Input types
                    ht.ListParam(ht.TypeTypeParam(ht.TypeBound.Any)),
                    # Output types
                    ht.ListParam(ht.TypeTypeParam(ht.TypeBound.Any)),
                ],
                body=ht.FunctionType(
                    input=[ht.RowVariable(1, ht.TypeBound.Any)],
                    output=[ht.RowVariable(2, ht.TypeBound.Any)],
                ),
            )
        ),
        description="An unsupported operation stub emitted by Guppy.",
    )
)


@dataclass
class UnsupportedOp(ops.AsExtOp):
    """An unsupported operation stub emitted by Guppy.

    args:
        op_name: The name of the unsupported operation.
        inputs: The input types of the operation.
        outputs: The output types of the operation.
    """

    op_name: str
    inputs: list[ht.Type]
    outputs: list[ht.Type]

    def op_def(self) -> he.OpDef:
        return UNSUPPORTED_OP_DEF

    def type_args(self) -> list[ht.TypeArg]:
        op_name = ht.StringArg(self.op_name)
        input_args = ht.SequenceArg([ht.TypeTypeArg(ty) for ty in self.inputs])
        output_args = ht.SequenceArg([ht.TypeTypeArg(ty) for ty in self.outputs])
        return [op_name, input_args, output_args]

    def cached_signature(self) -> ht.FunctionType | None:
        return ht.FunctionType(self.inputs, self.outputs)

    @classmethod
    def from_ext(cls, custom: ops.ExtOp) -> UnsupportedOp:
        match custom:
            case ops.ExtOp(_op_def=op_def, args=args):
                if op_def.qualified_name() == UNSUPPORTED_OP_DEF.qualified_name():
                    [op_name, input_args, output_args] = args
                    assert isinstance(op_name, ht.StringArg), (
                        "The first argument to a guppylang.unsupported op "
                        "must be the operation name"
                    )
                    op_name = op_name.value
                    return cls(
                        op_name=op_name,
                        inputs=[*_arg_seq_to_types(input_args)],
                        outputs=[*_arg_seq_to_types(output_args)],
                    )
        msg = f"Invalid custom op: {custom}"
        raise ops.AsExtOp.InvalidExtOp(msg)

    @property
    def num_out(self) -> int:
        return len(self.outputs)


def _arg_seq_to_types(args: ht.TypeArg) -> Iterator[ht.Type]:
    """Converts a SequenceArg of type arguments into a sequence of types."""
    assert isinstance(args, ht.SequenceArg)
    for arg in args.elems:
        assert isinstance(arg, ht.TypeTypeArg)
        yield arg.ty



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

import tests.error.import_errors.lib as lib

module = GuppyModule("test")
module.load(lib)


@guppy(module)
def main(x: "lib.MyType2[int]") -> int:
    return 0


module.compile()



"""Type checking code for top-level and nested function definitions.

For top-level functions, we take the `DefinedFunction` containing the `ast.FunctionDef`
node straight from the Python AST. We build a CFG, check it, and return a
`CheckedFunction` containing a `CheckedCFG` with type annotations.
"""

import ast
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from guppylang.ast_util import return_nodes_in_ast, with_loc
from guppylang.cfg.bb import BB
from guppylang.cfg.builder import CFGBuilder
from guppylang.checker.cfg_checker import CheckedCFG, check_cfg
from guppylang.checker.core import Context, Globals, Place, Variable
from guppylang.checker.errors.generic import UnsupportedError
from guppylang.definition.common import DefId
from guppylang.diagnostic import Error, Help, Note
from guppylang.error import GuppyError
from guppylang.nodes import CheckedNestedFunctionDef, NestedFunctionDef
from guppylang.tys.parsing import parse_function_io_types
from guppylang.tys.ty import FunctionType, InputFlags, NoneType

if TYPE_CHECKING:
    from guppylang.tys.param import Parameter


@dataclass(frozen=True)
class IllegalAssignError(Error):
    title: ClassVar[str] = "Illegal assignment"
    span_label: ClassVar[str] = (
        "Variable `{var}` may not be assigned to since `{var}` is captured from an "
        "outer scope"
    )
    var: str

    @dataclass(frozen=True)
    class DefHint(Note):
        span_label: ClassVar[str] = "`{var}` defined here"
        var: str


@dataclass(frozen=True)
class MissingArgAnnotationError(Error):
    title: ClassVar[str] = "Missing type annotation"
    span_label: ClassVar[str] = "Argument requires a type annotation"


@dataclass(frozen=True)
class MissingReturnAnnotationError(Error):
    title: ClassVar[str] = "Missing type annotation"
    span_label: ClassVar[str] = "Return type must be annotated"

    @dataclass(frozen=True)
    class ReturnNone(Help):
        message: ClassVar[str] = (
            "Looks like `{func}` doesn't return anything. Consider annotating it with "
            "`-> None`."
        )
        func: str


def check_global_func_def(
    func_def: ast.FunctionDef, ty: FunctionType, globals: Globals
) -> CheckedCFG[Place]:
    """Type checks a top-level function definition."""
    args = func_def.args.args
    returns_none = isinstance(ty.output, NoneType)
    assert ty.input_names is not None

    cfg = CFGBuilder().build(func_def.body, returns_none, globals)
    inputs = [
        Variable(x, inp.ty, loc, inp.flags)
        for x, inp, loc in zip(ty.input_names, ty.inputs, args, strict=True)
    ]
    generic_params = {
        param.name: param.with_idx(i) for i, param in enumerate(ty.params)
    }
    return check_cfg(cfg, inputs, ty.output, generic_params, func_def.name, globals)


def check_nested_func_def(
    func_def: NestedFunctionDef, bb: BB, ctx: Context
) -> CheckedNestedFunctionDef:
    """Type checks a local (nested) function definition."""
    func_ty = check_signature(func_def, ctx.globals)
    assert func_ty.input_names is not None

    if func_ty.parametrized:
        raise GuppyError(
            UnsupportedError(func_def, "Nested generic function definitions")
        )

    # We've already built the CFG for this function while building the CFG of the
    # enclosing function
    cfg = func_def.cfg

    # Find captured variables
    parent_cfg = bb.containing_cfg
    def_ass_before = set(func_ty.input_names) | ctx.locals.keys()
    maybe_ass_before = def_ass_before | parent_cfg.maybe_ass_before[bb]
    inout_vars = inout_var_names(func_ty)
    cfg.analyze(def_ass_before, maybe_ass_before, inout_vars)
    captured = {
        x: (ctx.locals[x], using_bb.vars.used[x])
        for x, using_bb in cfg.live_before[cfg.entry_bb].items()
        if x not in func_ty.input_names and x in ctx.locals
    }

    # Captured variables may never be assigned to
    for bb in cfg.bbs:
        for v, _ in captured.values():
            x = v.name
            if x in bb.vars.assigned:
                err = IllegalAssignError(bb.vars.assigned[x], x)
                err.add_sub_diagnostic(IllegalAssignError.DefHint(v.defined_at, x))
                raise GuppyError(err)

    # Construct inputs for checking the body CFG
    inputs = [v for v, _ in captured.values()] + [
        Variable(x, inp.ty, func_def.args.args[i], inp.flags)
        for i, (x, inp) in enumerate(
            zip(func_ty.input_names, func_ty.inputs, strict=True)
        )
    ]
    def_id = DefId.fresh()
    globals = ctx.globals

    # Check if the body contains a free (recursive) occurrence of the function name.
    # By checking if the name is free at the entry BB, we avoid false positives when
    # a user shadows the name with a local variable
    if func_def.name in cfg.live_before[cfg.entry_bb]:
        if not captured:
            # If there are no captured vars, we treat the function like a global name
            from guppylang.definition.function import ParsedFunctionDef

            func = ParsedFunctionDef(
                def_id, func_def.name, func_def, func_ty, globals.python_scope, None
            )
            globals = ctx.globals | Globals(
                {func.id: func}, {func_def.name: func.id}, {}, {}
            )
        else:
            # Otherwise, we treat it like a local name
            inputs.append(Variable(func_def.name, func_def.ty, func_def))

    checked_cfg = check_cfg(cfg, inputs, func_ty.output, {}, func_def.name, globals)
    checked_def = CheckedNestedFunctionDef(
        def_id,
        checked_cfg,
        func_ty,
        captured,
        name=func_def.name,
        args=func_def.args,
        body=func_def.body,
        decorator_list=func_def.decorator_list,
        returns=func_def.returns,
        type_comment=func_def.type_comment,
    )
    return with_loc(func_def, checked_def)


def check_signature(func_def: ast.FunctionDef, globals: Globals) -> FunctionType:
    """Checks the signature of a function definition and returns the corresponding
    Guppy type."""
    if len(func_def.args.posonlyargs) != 0:
        raise GuppyError(
            UnsupportedError(func_def.args.posonlyargs[0], "Positional-only parameters")
        )
    if len(func_def.args.kwonlyargs) != 0:
        raise GuppyError(
            UnsupportedError(func_def.args.kwonlyargs[0], "Keyword-only parameters")
        )
    if func_def.args.vararg is not None:
        raise GuppyError(UnsupportedError(func_def.args.vararg, "Variadic args"))
    if func_def.args.kwarg is not None:
        raise GuppyError(UnsupportedError(func_def.args.kwarg, "Keyword args"))
    if func_def.args.defaults:
        raise GuppyError(
            UnsupportedError(func_def.args.defaults[0], "Default arguments")
        )
    if func_def.returns is None:
        err = MissingReturnAnnotationError(func_def)
        # TODO: Error location is incorrect
        if all(r.value is None for r in return_nodes_in_ast(func_def)):
            err.add_sub_diagnostic(
                MissingReturnAnnotationError.ReturnNone(None, func_def.name)
            )
        raise GuppyError(err)

    # TODO: Prepopulate mapping when using Python 3.12 style generic functions
    param_var_mapping: dict[str, Parameter] = {}
    input_nodes = []
    input_names = []
    for inp in func_def.args.args:
        ty_ast = inp.annotation
        if ty_ast is None:
            raise GuppyError(MissingArgAnnotationError(inp))
        input_nodes.append(ty_ast)
        input_names.append(inp.arg)
    inputs, output = parse_function_io_types(
        input_nodes, func_def.returns, func_def, globals, param_var_mapping, True
    )
    return FunctionType(
        inputs,
        output,
        input_names,
        sorted(param_var_mapping.values(), key=lambda v: v.idx),
    )


def parse_function_with_docstring(
    func_ast: ast.FunctionDef,
) -> tuple[ast.FunctionDef, str | None]:
    """Check if the first line of a function is a docstring.

    If it is, return the function with the docstring removed, plus the docstring.
    Else, return the original function and `None`
    """
    docstring = None
    match func_ast.body:
        case [doc, *xs]:
            if (
                isinstance(doc, ast.Expr)
                and isinstance(doc.value, ast.Constant)
                and isinstance(doc.value.value, str)
            ):
                docstring = doc.value.value
                func_ast.body = xs
        case _:
            pass
    return func_ast, docstring


def inout_var_names(func_ty: FunctionType) -> list[str]:
    """Returns the names of all borrowed arguments in a function type."""
    assert func_ty.input_names is not None
    return [
        x
        for inp, x in zip(func_ty.inputs, func_ty.input_names, strict=True)
        if InputFlags.Inout in inp.flags
    ]



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit, measure


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class MyStruct1:
    x: "MyStruct2"


@guppy.struct(module)
class MyStruct2:
    q1: qubit
    q2: qubit


@guppy(module)
def foo(s: MyStruct1 @owned) -> MyStruct1:
    measure(s.x.q1)
    return MyStruct1(s.x)


module.compile()



from tests.util import compile_guppy


@compile_guppy
def foo(b: bool) -> int:
    if b:
        def bar() -> int:
            return 0
    else:
        def bar() -> bool:
            return False

    return bar()



from hugr import Hugr
from hugr.package import Package, PackagePointer

from pathlib import Path
import pytest
import subprocess
from typing import Any


@pytest.fixture(scope="session")
def export_test_cases_dir(request):
    r = request.config.getoption("--export-test-cases")
    if r and not r.exists():
        r.mkdir(parents=True)
    return r


def get_validator() -> Path | None:
    """
    Returns the path to the `validator` binary, if it exists.
    Otherwise, returns `None`.
    """
    bin_path = Path(__file__).parent.parent.parent / "target" / "release" / "validator"
    if bin_path.exists():
        return bin_path

    return None


@pytest.fixture(scope="session")
def validate(request, export_test_cases_dir: Path):
    if request.config.getoption("validation"):
        # Check if the validator is installed
        validator = get_validator()
        if validator is None:
            pytest.fail("Run `cargo build -p release` to install the validator")
    else:
        pytest.skip("Skipping validation tests as requested")

    def validate_bytes(package_bytes: bytes):
        # Executes `cargo run -p validator -- validate -`
        # passing the hugr JSON as stdin
        p = subprocess.run(  # noqa: S603
            [validator, "validate", "-"],
            text=False,
            input=package_bytes,
            capture_output=True,
        )

        if p.returncode != 0:
            raise RuntimeError(f"{p.stderr}")

    def validate_impl(package: Package | PackagePointer | Hugr, name=None):
        if isinstance(package, PackagePointer):
            package = package.package
        if isinstance(package, Hugr):
            package = Package([package])
        # Validate via the json encoding
        package_bytes = package.to_bytes()

        if export_test_cases_dir:
            file_name = f"{request.node.name}{f'_{name}' if name else ''}.hugr"
            export_file = export_test_cases_dir / file_name
            export_file.write_bytes(package_bytes)

        validate_bytes(package_bytes)

    return validate_impl


class LLVMException(Exception):
    pass


def _run_fn(run_fn_name: str):
    def f(module: PackagePointer, expected: Any, fn_name: str = "main"):
        try:
            import execute_llvm

            fn = getattr(execute_llvm, run_fn_name)
            if not fn:
                pytest.skip("Skipping llvm execution")

            package_bytes = module.package.to_bytes()
            res = fn(package_bytes, fn_name)
            if res != expected:
                raise LLVMException(
                    f"Expected value ({expected}) doesn't match actual value ({res})"
                )
        except AttributeError:
            pytest.skip("Skipping llvm execution")
        except ImportError:
            pytest.skip("Skipping llvm execution")

    return f


@pytest.fixture
def run_int_fn():
    return _run_fn("run_int_function")


@pytest.fixture
def run_float_fn_approx():
    """Like run_int_fn, but takes optional additional parameters `rel`, `abs` and `nan_ok`
    as per `pytest.approx`."""
    run_fn = _run_fn("run_float_function")

    def run_approx(
        hugr: PackagePointer,
        expected: float,
        fn_name: str = "main",
        *,
        rel: float | None = None,
        abs: float | None = None,
        nan_ok: bool = False,
    ):
        return run_fn(hugr, pytest.approx(expected, rel=rel, abs=abs, nan_ok=nan_ok))

    return run_approx



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module1 = GuppyModule("test1")


@guppy(module1)
def foo(x: float) -> int:
    return x


module2 = GuppyModule("test2")

# Load invokes `check` on the imported module, so this will fail even though we don't
# compile `module2`
module2.load(foo)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.tys.ty import NoneType


module = GuppyModule("test")


@guppy.type(NoneType().to_hugr(), module=module)
class MyIter:
    """An iterator where the `__next__` method has the wrong signature."""

    @guppy.declare(module)
    def __next__(self: "MyIter") -> tuple["MyIter", float]:
        ...

    @guppy.declare(module)
    def __hasnext__(self: "MyIter") -> tuple[bool, "MyIter"]:
        ...

    @guppy.declare(module)
    def __end__(self: "MyIter") -> None:
        ...


@guppy.type(NoneType().to_hugr(), module=module)
class MyType:
    """Type that produces the iterator above."""

    @guppy.declare(module)
    def __iter__(self: "MyType") -> MyIter:
        ...


@guppy(module)
def test(x: MyType) -> None:
    for _ in x:
        pass


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)

@guppy(module)
def main(xs: array[qubit, 2] @ owned) -> array[int, 2]:
   ys = copy(xs)
   return ys

module.compile()



import pytest

from tests.integration.util import functional, _
from tests.util import compile_guppy


@pytest.mark.skip
def test_if_no_else(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        _ @ functional
        if x:
            y += 1
        return y

    validate(foo)


@pytest.mark.skip
def test_if_else(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        _ @ functional
        if x:
            y += 1
        else:
            y -= 1
        return y

    validate(foo)


@pytest.mark.skip
def test_if_elif(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        _ @ functional
        if x:
            y += 1
        elif y > 4:
            y *= 7
        return y

    validate(foo)


@pytest.mark.skip
def test_if_elif_else(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> int:
        _ @ functional
        if x:
            y += 1
        elif y > 4:
            y *= 7
        else:
            y = 1337
        return y

    validate(foo)


@pytest.mark.skip
def test_infinite_loop(validate):
    @compile_guppy
    def foo() -> int:
        while True:
            pass
        return 0

    validate(foo)


@pytest.mark.skip
def test_counting_loop(validate):
    @compile_guppy
    def foo(i: int) -> int:
        while i > 0:
            i -= 1
        return 0

    validate(foo)


@pytest.mark.skip
def test_nested_loop(validate):
    @compile_guppy
    def foo(x: int, y: int) -> int:
        p = 0
        _ @ functional
        while x > 0:
            s = 0
            while y > 0:
                s += x
                y -= 1
            p += s
            x -= 1
        return p

    validate(foo)



from typing import Generic, TYPE_CHECKING

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


if TYPE_CHECKING:
    from collections.abc import Callable


def test_basic_defs(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class EmptyStruct:
        pass

    @guppy.struct(module)
    class OneMemberStruct:
        x: int

    @guppy.struct(module)
    class TwoMemberStruct:
        x: tuple[bool, int]
        y: float

    @guppy.struct(module)
    class DocstringStruct:
        """This is struct with a docstring!"""

        x: int

    @guppy(module)
    def main(
        a: EmptyStruct, b: OneMemberStruct, c: TwoMemberStruct, d: DocstringStruct
    ) -> None:
        EmptyStruct()
        OneMemberStruct(42)
        TwoMemberStruct((True, 0), 1.0)
        DocstringStruct(-1)

    validate(module.compile())


def test_backward_ref(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class StructA:
        x: int

    @guppy.struct(module)
    class StructB:
        y: StructA

    @guppy(module)
    def main(a: StructA, b: StructB) -> None:
        StructB(a)

    validate(module.compile())


def test_forward_ref(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class StructA:
        x: "StructB"

    @guppy.struct(module)
    class StructB:
        y: int

    @guppy(module)
    def main(a: StructA, b: StructB) -> None:
        StructA(b)

    validate(module.compile())


def test_generic(validate):
    module = GuppyModule("module")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy.struct(module)
    class StructA(Generic[T]):
        x: tuple[int, T]

    @guppy.struct(module)
    class StructC:
        a: StructA[int]
        b: StructA[list[bool]]
        c: "StructB[float, StructB[bool, int]]"

    @guppy.struct(module)
    class StructB(Generic[S, T]):
        x: S
        y: StructA[T]

    @guppy(module)
    def main(a: StructA[StructA[float]], b: StructB[bool, int], c: StructC) -> None:
        x = StructA((0, False))
        y = StructA((0, -5))
        StructA((0, x))
        StructB(x, a)
        StructC(y, StructA((0, [])), StructB(42.0, StructA((4, b))))

    validate(module.compile())


def test_methods(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class StructA:
        x: int

        @guppy(module)
        def foo(self: "StructA", y: int) -> int:
            return 2 * self.x + y

    @guppy.struct(module)
    class StructB:
        x: int
        y: float

        @guppy(module)
        def bar(self: "StructB", a: StructA) -> float:
            return a.foo(self.x) + self.y

    @guppy(module)
    def main(a: StructA, b: StructB) -> tuple[int, float]:
        return a.foo(1), b.bar(a)

    validate(module.compile())


def test_higher_order(validate):
    module = GuppyModule("module")
    T = guppy.type_var("T", module=module)

    @guppy.struct(module)
    class Struct(Generic[T]):
        x: T

    @guppy(module)
    def factory(mk_struct: "Callable[[int], Struct[int]]", x: int) -> Struct[int]:
        return mk_struct(x)

    @guppy(module)
    def main() -> None:
        factory(Struct, 42)

    validate(module.compile())


def test_wiring(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class MyStruct:
        x: int

    @guppy(module)
    def foo() -> MyStruct:
        s = 0
        # This tests that reassigning `s` invalidates the old `s = 0` wire when
        # compiling to Hugr.
        s = MyStruct(42)
        return s

    validate(module.compile())


def test_field_access_and_drop(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class MyStruct:
        x: int
        y: float
        z: tuple[int, int]

    @guppy(module)
    def foo() -> float:
        # Access a field of an unnamed struct,
        # dropping all the other fields
        return MyStruct(42, 3.14, (1, 2)).y

    validate(module.compile())



import ast
from dataclasses import dataclass, field

from hugr import Node, Wire, val
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang.ast_util import AstNode
from guppylang.checker.core import Globals
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.definition.common import CompilableDef, ParsableDef
from guppylang.definition.value import CompiledValueDef, ValueDef
from guppylang.span import SourceMap
from guppylang.tys.parsing import type_from_ast


@dataclass(frozen=True)
class RawExternDef(ParsableDef):
    """A raw extern symbol definition provided by the user."""

    symbol: str
    constant: bool
    type_ast: ast.expr

    description: str = field(default="extern", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ExternDef":
        """Parses and checks the user-provided signature of the function."""
        return ExternDef(
            self.id,
            self.name,
            self.defined_at,
            type_from_ast(self.type_ast, globals, {}),
            self.symbol,
            self.constant,
            self.type_ast,
        )


@dataclass(frozen=True)
class ExternDef(RawExternDef, ValueDef, CompilableDef):
    """An extern symbol definition."""

    def compile_outer(self, graph: DefinitionBuilder[OpVar]) -> "CompiledExternDef":
        """Adds a Hugr constant node for the extern definition to the provided graph."""
        # The `typ` field must be serialized at this point, to ensure that the
        # `Extension` is serializable.
        custom_const = {
            "symbol": self.symbol,
            "typ": self.ty.to_hugr()._to_serial_root(),
            "constant": self.constant,
        }
        value = val.Extension(
            name="ConstExternalSymbol",
            typ=self.ty.to_hugr(),
            val=custom_const,
            extensions=["prelude"],
        )
        const_node = graph.add_const(value)
        return CompiledExternDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.symbol,
            self.constant,
            self.type_ast,
            const_node,
        )


@dataclass(frozen=True)
class CompiledExternDef(ExternDef, CompiledValueDef):
    """An extern symbol definition that has been compiled to a Hugr constant."""

    const_node: Node

    def load(self, dfg: DFContainer, ctx: CompilerContext, node: AstNode) -> Wire:
        """Loads the extern value into a local Hugr dataflow graph."""
        return dfg.builder.load(self.const_node)



"""Custom AST nodes used by Guppy"""

import ast
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from guppylang.ast_util import AstNode
from guppylang.tys.const import Const
from guppylang.tys.subst import Inst
from guppylang.tys.ty import FunctionType, StructType, Type

if TYPE_CHECKING:
    from guppylang.cfg.cfg import CFG
    from guppylang.checker.cfg_checker import CheckedCFG
    from guppylang.checker.core import Place, Variable
    from guppylang.definition.common import DefId
    from guppylang.definition.struct import StructField
    from guppylang.tys.param import ConstParam


class PlaceNode(ast.expr):
    place: "Place"

    _fields = ("place",)


class GlobalName(ast.Name):
    id: str
    def_id: "DefId"

    _fields = (
        "id",
        "def_id",
    )


class GenericParamValue(ast.Name):
    id: str
    param: "ConstParam"

    _fields = (
        "id",
        "param",
    )


class LocalCall(ast.expr):
    func: ast.expr
    args: list[ast.expr]

    _fields = (
        "func",
        "args",
    )


class GlobalCall(ast.expr):
    def_id: "DefId"
    args: list[ast.expr]
    type_args: Inst  # Inferred type arguments

    _fields = (
        "def_id",
        "args",
        "type_args",
    )


class TensorCall(ast.expr):
    """A call to a tuple of functions. Behaves like a local call, but more
    unpacking of tuples is required at compilation"""

    func: ast.expr
    args: list[ast.expr]
    tensor_ty: FunctionType

    _fields = (
        "func",
        "args",
        "tensor_ty",
    )


class TypeApply(ast.expr):
    value: ast.expr
    inst: Inst

    _fields = (
        "value",
        "inst",
    )


class PartialApply(ast.expr):
    """A partial function application.

    This node is emitted when methods are loaded as values, since this requires
    partially applying the `self` argument.
    """

    func: ast.expr
    args: list[ast.expr]

    _fields = (
        "func",
        "args",
    )


class FieldAccessAndDrop(ast.expr):
    """A field access on a struct, dropping all the remaining other fields."""

    value: ast.expr
    struct_ty: "StructType"
    field: "StructField"

    _fields = (
        "value",
        "struct_ty",
        "field",
    )


class SubscriptAccessAndDrop(ast.expr):
    """A subscript element access on an object, dropping all the remaining items."""

    item: "Variable"
    item_expr: ast.expr
    getitem_expr: ast.expr
    original_expr: ast.Subscript

    _fields = ("item", "item_expr", "getitem_expr", "original_expr")


class MakeIter(ast.expr):
    """Creates an iterator using the `__iter__` magic method.

    This node is inserted in `for` loops and list comprehensions.
    """

    value: ast.expr
    unwrap_size_hint: bool

    # Node that triggered the creation of this iterator. For example, a for loop stmt.
    # It is not mentioned in `_fields` so that it is not visible to AST visitors
    origin_node: ast.AST

    _fields = ("value",)

    def __init__(
        self, value: ast.expr, origin_node: ast.AST, unwrap_size_hint: bool = True
    ) -> None:
        super().__init__(value)
        self.origin_node = origin_node
        self.unwrap_size_hint = unwrap_size_hint


class IterHasNext(ast.expr):
    """Checks if an iterator has a next element using the `__hasnext__` magic method.

    This node is inserted in `for` loops and list comprehensions.
    """

    value: ast.expr

    _fields = ("value",)


class IterNext(ast.expr):
    """Obtains the next element of an iterator using the `__next__` magic method.

    This node is inserted in `for` loops and list comprehensions.
    """

    value: ast.expr

    _fields = ("value",)


class IterEnd(ast.expr):
    """Finalises an iterator using the `__end__` magic method.

    This node is inserted in `for` loops and list comprehensions. It is needed to
    consume linear iterators once they are finished.
    """

    value: ast.expr

    _fields = ("value",)


class DesugaredGenerator(ast.expr):
    """A single desugared generator in a list comprehension.

    Stores assignments of the original generator targets as well as dummy variables for
    the iterator and hasnext test.
    """

    iter_assign: ast.Assign
    next_call: ast.expr
    iter: ast.expr
    target: ast.expr
    ifs: list[ast.expr]

    borrowed_outer_places: "list[Place]"

    _fields = (
        "iter_assign",
        "next_call",
        "iter",
        "target",
        "ifs",
    )


class DesugaredGeneratorExpr(ast.expr):
    """A desugared generator expression."""

    elt: ast.expr
    generators: list[DesugaredGenerator]

    _fields = (
        "elt",
        "generators",
    )


class DesugaredListComp(ast.expr):
    """A desugared list comprehension."""

    elt: ast.expr
    generators: list[DesugaredGenerator]

    _fields = (
        "elt",
        "generators",
    )


class DesugaredArrayComp(ast.expr):
    """A desugared array comprehension."""

    elt: ast.expr
    generator: DesugaredGenerator
    length: Const
    elt_ty: Type

    _fields = (
        "elt",
        "generator",
        "length",
        "elt_ty",
    )


class PyExpr(ast.expr):
    """A compile-time evaluated `py(...)` expression."""

    value: ast.expr

    _fields = ("value",)


class ResultExpr(ast.expr):
    """A `result(tag, value)` expression."""

    value: ast.expr
    base_ty: Type
    #: Array length in case this is an array result, otherwise `None`
    array_len: Const | None
    tag: str

    _fields = ("value", "base_ty", "array_len", "tag")


class PanicExpr(ast.expr):
    """A `panic(msg, *args)` expression."""

    msg: str
    values: list[ast.expr]

    _fields = ("msg", "values")


class BarrierExpr(ast.expr):
    """A `barrier(*args)` expression."""

    args: list[ast.expr]
    func_ty: FunctionType
    _fields = ("args", "func_ty")


AnyCall = LocalCall | GlobalCall | TensorCall | BarrierExpr


class InoutReturnSentinel(ast.expr):
    """An invisible expression corresponding to an implicit use of borrowed vars
    whenever a function returns."""

    var: "Place | str"

    _fields = ("var",)


class UnpackPattern(ast.expr):
    """The LHS of an unpacking assignment like `a, *bs, c = ...` or
    `[a, *bs, c] = ...`."""

    #: Patterns occurring on the left of the starred target
    left: list[ast.expr]

    #: The starred target or `None` if there is none
    starred: ast.expr | None

    #: Patterns occurring on the right of the starred target. This will be an empty list
    #: if there is no starred target
    right: list[ast.expr]

    _fields = ("left", "starred", "right")


class TupleUnpack(ast.expr):
    """The LHS of an unpacking assignment of a tuple."""

    #: The (possibly starred) unpacking pattern
    pattern: UnpackPattern

    _fields = ("pattern",)


class IterableUnpack(ast.expr):
    """The LHS of an unpacking assignment of an iterable type."""

    #: The (possibly starred) unpacking pattern
    pattern: UnpackPattern

    #: Comprehension that collects the RHS iterable into an array
    compr: DesugaredArrayComp

    #: Dummy variable that the RHS should be bound to. This variable is referenced in
    #: `compr`
    rhs_var: PlaceNode

    # Don't mention the comprehension in _fields to avoid visitors recursing it
    _fields = ("pattern",)

    def __init__(
        self, pattern: UnpackPattern, compr: DesugaredArrayComp, rhs_var: PlaceNode
    ) -> None:
        super().__init__(pattern)
        self.compr = compr
        self.rhs_var = rhs_var


#: Any unpacking operation.
AnyUnpack = TupleUnpack | IterableUnpack


class NestedFunctionDef(ast.FunctionDef):
    cfg: "CFG"
    ty: FunctionType
    docstring: str | None

    def __init__(self, cfg: "CFG", ty: FunctionType, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.cfg = cfg
        self.ty = ty


class CheckedNestedFunctionDef(ast.FunctionDef):
    def_id: "DefId"
    cfg: "CheckedCFG[Place]"
    ty: FunctionType

    #: Mapping from names to variables captured by this function, together with an AST
    #: node witnessing a use of the captured variable in the function body.
    captured: Mapping[str, tuple["Variable", AstNode]]

    def __init__(
        self,
        def_id: "DefId",
        cfg: "CheckedCFG[Place]",
        ty: FunctionType,
        captured: Mapping[str, tuple["Variable", AstNode]],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.def_id = def_id
        self.cfg = cfg
        self.ty = ty
        self.captured = captured



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from tests.util import compile_guppy


def test_call(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo() -> int:
        return 42

    @guppy(module)
    def bar() -> int:
        return foo()

    validate(module.compile())


def test_call_back(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> int:
        return bar(x)

    @guppy(module)
    def bar(x: int) -> int:
        return x

    validate(module.compile())


def test_recursion(validate):
    @compile_guppy
    def main(x: int) -> int:
        return main(x)

    validate(main)


def test_mutual_recursion(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> int:
        return bar(x)

    @guppy(module)
    def bar(x: int) -> int:
        return foo(x)

    validate(module.compile())


def test_unary_tuple(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> tuple[int]:
        return (x,)

    @guppy(module)
    def bar(x: int) -> int:
        (y,) = foo(x)
        return y

    validate(module.compile())


def test_method_call(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> int:
        return x.__add__(2)

    validate(module.compile())



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array

module = GuppyModule("test")

n = guppy.nat_var("n", module=module)


@guppy.declare(module)
def foo(x: array[int, n], y: array[int, n]) -> None:
    ...


@guppy(module)
def main(x: array[int, 42], y: array[int, 43]) -> None:
    foo(x, y)


module.compile()



import ast
from dataclasses import dataclass, field
from typing import Any, cast

import hugr.build.function as hf
from hugr import Hugr, Wire, ops, val
from hugr import tys as ht
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang.ast_util import AstNode, has_empty_body, with_loc
from guppylang.checker.core import Context, Globals, PyScope
from guppylang.checker.errors.py_errors import (
    PytketSignatureMismatch,
    Tket2NotInstalled,
)
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import check_signature
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.definition.common import (
    CompilableDef,
    ParsableDef,
)
from guppylang.definition.declaration import BodyNotEmptyError
from guppylang.definition.function import (
    PyFunc,
    compile_call,
    load_with_args,
    parse_py_func,
)
from guppylang.definition.ty import TypeDef
from guppylang.definition.value import CallableDef, CallReturnWires, CompiledCallableDef
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.nodes import GlobalCall
from guppylang.span import SourceMap, Span, ToSpan
from guppylang.std._internal.compiler.array import (
    array_discard_empty,
    array_new,
    array_pop,
)
from guppylang.std._internal.compiler.prelude import build_unwrap
from guppylang.tys.builtin import array_type, bool_type
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    Type,
    row_to_type,
)


@dataclass(frozen=True)
class RawPytketDef(ParsableDef):
    """A raw function stub definition describing the signature of a circuit.

    Args:
        id: The unique definition identifier.
        name: The name of the function stub.
        defined_at: The AST node where the stub was defined.
        python_func: The Python function stub.
        python_scope: The Python scope where the function stub was defined.
        input_circuit: The user-provided pytket circuit.
    """

    python_func: PyFunc
    python_scope: PyScope = field(repr=False)
    input_circuit: Any

    description: str = field(default="pytket circuit", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ParsedPytketDef":
        """Parses and checks the user-provided signature matches the user-provided
        circuit.
        """
        # Retrieve stub signature.
        func_ast, _ = parse_py_func(self.python_func, sources)
        if not has_empty_body(func_ast):
            # Function stub should have empty body.
            raise GuppyError(BodyNotEmptyError(func_ast.body[0], self.name))
        stub_signature = check_signature(
            func_ast, globals.with_python_scope(self.python_scope)
        )

        # Compare signatures.
        circuit_signature = _signature_from_circuit(
            self.input_circuit, globals, self.defined_at
        )
        if not (
            circuit_signature.inputs == stub_signature.inputs
            and circuit_signature.output == stub_signature.output
        ):
            err = PytketSignatureMismatch(func_ast, self.name)
            err.add_sub_diagnostic(
                PytketSignatureMismatch.TypeHint(None, circ_sig=circuit_signature)
            )
            raise GuppyError(err)
        return ParsedPytketDef(
            self.id, self.name, func_ast, stub_signature, self.input_circuit, False
        )


@dataclass(frozen=True)
class RawLoadPytketDef(ParsableDef):
    """A raw definition for loading pytket circuits without explicit function stub.

    Args:
        id: The unique definition identifier.
        name: The name of the circuit function.
        defined_at: The AST node of the definition (here always None).
        source_span: The source span where the circuit was loaded.
        input_circuit: The user-provided pytket circuit.
        use_arrays: Whether the circuit function should use arrays as input types.
    """

    source_span: Span | None
    input_circuit: Any
    use_arrays: bool

    description: str = field(default="pytket circuit", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ParsedPytketDef":
        """Creates a function signature based on the user-provided circuit."""
        circuit_signature = _signature_from_circuit(
            self.input_circuit, globals, self.source_span, self.use_arrays
        )

        return ParsedPytketDef(
            self.id,
            self.name,
            self.defined_at,
            circuit_signature,
            self.input_circuit,
            self.use_arrays,
        )


@dataclass(frozen=True)
class ParsedPytketDef(CallableDef, CompilableDef):
    """A circuit definition with signature.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node of the function stub, if there is one.
        ty: The type of the function.
        input_circuit: The user-provided pytket circuit.
        use_arrays: Whether the circuit function should use arrays as input types.
    """

    ty: FunctionType
    input_circuit: Any
    use_arrays: bool

    description: str = field(default="pytket circuit", init=False)

    def compile_outer(self, module: DefinitionBuilder[OpVar]) -> "CompiledPytketDef":
        """Adds a Hugr `FuncDefn` node for this function to the Hugr."""
        try:
            import pytket

            if isinstance(self.input_circuit, pytket.circuit.Circuit):
                from tket2.circuit import (  # type: ignore[import-untyped, import-not-found, unused-ignore]
                    Tk2Circuit,
                )

                circ = Hugr.load_json(Tk2Circuit(self.input_circuit).to_hugr_json())  # type: ignore[attr-defined, unused-ignore]
                mapping = module.hugr.insert_hugr(circ)
                hugr_func = mapping[circ.root]

                func_type = self.ty.to_hugr_poly()
                outer_func = module.define_function(
                    self.name, func_type.body.input, func_type.body.output
                )

                # Initialise every input bit in the circuit as false.
                # TODO: Provide the option for the user to pass this input as well.
                bool_wires = [
                    outer_func.load(val.FALSE) for _ in range(self.input_circuit.n_bits)
                ]

                input_list = []
                if self.use_arrays:
                    # If the input is given as arrays, we need to unpack each element in
                    # them into separate wires.
                    # TODO: Replace with actual unpack HUGR op once
                    # https://github.com/CQCL/hugr/issues/1947 is done.
                    def unpack(
                        array: Wire, elem_ty: ht.Type, length: int
                    ) -> list[Wire]:
                        err = "Internal error: unpacking of array failed"
                        elts: list[Wire] = []
                        for i in range(length):
                            res = outer_func.add_op(
                                array_pop(elem_ty, length - i, True), array
                            )
                            [elt_opt, array] = build_unwrap(outer_func, res, err)
                            [elt] = build_unwrap(outer_func, elt_opt, err)
                            elts.append(elt)
                        outer_func.add_op(array_discard_empty(elem_ty), array)
                        return elts

                    # Must be same length due to earlier signature computation /
                    # comparison.
                    for q_reg, wire in zip(
                        self.input_circuit.q_registers,
                        list(outer_func.inputs()),
                        strict=True,
                    ):
                        input_list.extend(unpack(wire, ht.Option(ht.Qubit), q_reg.size))

                else:
                    # Otherwise pass inputs directly.
                    input_list = list(outer_func.inputs())

                call_node = outer_func.call(hugr_func, *(input_list + bool_wires))

                # Pytket circuit hugr has qubit and bool wires in the opposite
                # order to Guppy output wires.
                output_list: list[Wire] = list(call_node.outputs())
                wires = (
                    output_list[self.input_circuit.n_qubits :]
                    + output_list[: self.input_circuit.n_qubits]
                )

                if self.use_arrays:

                    def pack(elems: list[Wire], elem_ty: ht.Type, length: int) -> Wire:
                        elem_opts = [
                            outer_func.add_op(ops.Some(elem_ty), elem) for elem in elems
                        ]
                        return outer_func.add_op(
                            array_new(ht.Option(elem_ty), length), *elem_opts
                        )

                    array_wires: list[Wire] = []
                    wire_idx = 0
                    # First pack bool results into an array.
                    for c_reg in self.input_circuit.c_registers:
                        array_wires.append(
                            pack(
                                wires[wire_idx : wire_idx + c_reg.size],
                                ht.Bool,
                                c_reg.size,
                            )
                        )
                        wire_idx = wire_idx + c_reg.size
                    # Then the borrowed qubits also need to be put back into arrays.
                    for q_reg in self.input_circuit.q_registers:
                        array_wires.append(
                            pack(
                                wires[wire_idx : wire_idx + q_reg.size],
                                ht.Qubit,
                                q_reg.size,
                            )
                        )
                        wire_idx = wire_idx + q_reg.size
                    wires = array_wires

                outer_func.set_outputs(*wires)

        except ImportError:
            pass

        return CompiledPytketDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.input_circuit,
            self.use_arrays,
            outer_func,
        )

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call."""
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, ty


@dataclass(frozen=True)
class CompiledPytketDef(ParsedPytketDef, CompiledCallableDef):
    """A function definition with a corresponding Hugr node.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        ty: The type of the function.
        input_circuit: The user-provided pytket circuit.
        func_df: The Hugr function definition.
        use_arrays: Whether the circuit function uses arrays as input types.
    """

    func_def: hf.Function

    def load_with_args(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the function as a value into a local Hugr dataflow graph."""
        # Use implementation from function definition.
        return load_with_args(type_args, dfg, self.ty, self.func_def)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        # Use implementation from function definition.
        return compile_call(args, type_args, dfg, self.ty, self.func_def)


def _signature_from_circuit(
    input_circuit: Any,
    globals: Globals,
    defined_at: ToSpan | None,
    use_arrays: bool = False,
) -> FunctionType:
    """Helper function for inferring a function signature from a pytket circuit."""
    try:
        import pytket

        if isinstance(input_circuit, pytket.circuit.Circuit):
            try:
                import tket2  # type: ignore[import-untyped, import-not-found, unused-ignore]  # noqa: F401

                qubit = cast(TypeDef, globals["qubit"]).check_instantiate([], globals)

                if use_arrays:
                    inputs = [
                        FuncInput(array_type(qubit, q_reg.size), InputFlags.Inout)
                        for q_reg in input_circuit.q_registers
                    ]
                    outputs = [
                        array_type(bool_type(), c_reg.size)
                        for c_reg in input_circuit.c_registers
                    ]
                    circuit_signature = FunctionType(
                        inputs,
                        row_to_type(outputs),
                    )
                else:
                    circuit_signature = FunctionType(
                        [FuncInput(qubit, InputFlags.Inout)] * input_circuit.n_qubits,
                        row_to_type([bool_type()] * input_circuit.n_bits),
                    )
            except ImportError:
                err = Tket2NotInstalled(defined_at)
                err.add_sub_diagnostic(Tket2NotInstalled.InstallInstruction(None))
                raise GuppyError(err) from None
        else:
            pass
    except ImportError:
        raise InternalGuppyError(
            "Pytket error should have been caught earlier"
        ) from None
    else:
        return circuit_signature



"""Compilers building list functions on top of hugr standard operations, that involve
multiple nodes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import hugr.std.collections.list
from hugr import Wire, ops
from hugr import tys as ht
from hugr.std.collections.list import List, ListVal

from guppylang.definition.custom import CustomCallCompiler
from guppylang.definition.value import CallReturnWires
from guppylang.error import InternalGuppyError
from guppylang.std._internal.compiler.arithmetic import (
    convert_ifromusize,
    convert_itousize,
)
from guppylang.std._internal.compiler.prelude import (
    build_unwrap,
    build_unwrap_left,
    build_unwrap_right,
)
from guppylang.tys.arg import TypeArg

if TYPE_CHECKING:
    from hugr.build.dfg import DfBase


# ------------------------------------------------------
# --------- std.collections operations -----------------
# ------------------------------------------------------


def _instantiate_list_op(
    name: str, elem_type: ht.Type, inp: list[ht.Type], out: list[ht.Type]
) -> ops.ExtOp:
    op_def = hugr.std.collections.list.EXTENSION.get_op(name)
    return ops.ExtOp(
        op_def,
        ht.FunctionType(inp, out),
        [ht.TypeTypeArg(elem_type)],
    )


def list_pop(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `pop` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "pop", elem_type, [list_type], [list_type, ht.Option(elem_type)]
    )


def list_push(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `push` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op("push", elem_type, [list_type, elem_type], [list_type])


def list_get(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `get` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "get", elem_type, [list_type, ht.USize()], [ht.Option(elem_type)]
    )


def list_set(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `set` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "set",
        elem_type,
        [list_type, ht.USize(), elem_type],
        [list_type, ht.Either([elem_type], [elem_type])],
    )


def list_insert(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `insert` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "insert",
        elem_type,
        [list_type, ht.USize(), elem_type],
        [list_type, ht.Either([elem_type], [ht.Unit])],
    )


def list_length(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `length` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "length", elem_type, [list_type], [list_type, ht.USize()]
    )


# ------------------------------------------------------
# --------- Custom compilers for non-native ops --------
# ------------------------------------------------------


class ListGetitemCompiler(CustomCallCompiler):
    """Compiler for the `list.__getitem__` function."""

    def build_classical_getitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.__getitem__` for classical lists."""
        idx = self.builder.add_op(convert_itousize(), idx)
        result = self.builder.add_op(list_get(elem_ty), list_wire, idx)
        elem = build_unwrap(self.builder, result, "List index out of bounds")
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def build_linear_getitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.__getitem__` for linear lists."""
        # Swap out the element at the given index with `None`. The `to_hugr`
        # implementation of the list type ensures that linear element types are turned
        # into optionals.
        elem_opt_ty = ht.Option(elem_ty)
        none = self.builder.add_op(ops.Tag(0, elem_opt_ty))
        idx = self.builder.add_op(convert_itousize(), idx)
        list_wire, result = self.builder.add_op(
            list_set(elem_opt_ty), list_wire, idx, none
        )
        elem_opt = build_unwrap_right(self.builder, result, "List index out of bounds")
        elem = build_unwrap(
            self.builder, elem_opt, "Linear list element has already been used"
        )
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire, idx] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_getitem(list_wire, idx, elem_ty_arg.ty.to_hugr())
        else:
            return self.build_classical_getitem(
                list_wire, idx, elem_ty_arg.ty.to_hugr()
            )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListSetitemCompiler(CustomCallCompiler):
    """Compiler for the `list.__setitem__` function."""

    def build_classical_setitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.__setitem__` for classical lists."""
        idx = self.builder.add_op(convert_itousize(), idx)
        list_wire, result = self.builder.add_op(list_set(elem_ty), list_wire, idx, elem)
        # Unwrap the result, but we don't have to hold onto the returned old value
        build_unwrap_right(self.builder, result, "List index out of bounds")
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def build_linear_setitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `array.__setitem__` for linear arrays."""
        # Embed the element into an optional
        elem_opt_ty = ht.Option(elem_ty)
        elem = self.builder.add_op(ops.Some(elem_ty), elem)
        idx = self.builder.add_op(convert_itousize(), idx)
        list_wire, result = self.builder.add_op(
            list_set(elem_opt_ty), list_wire, idx, elem
        )
        old_elem_opt = build_unwrap_right(
            self.builder, result, "List index out of bounds"
        )
        # Check that the old element was `None`
        build_unwrap_left(
            self.builder, old_elem_opt, "Linear list element has not been used"
        )
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire, idx, elem] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_setitem(
                list_wire, idx, elem, elem_ty_arg.ty.to_hugr()
            )
        else:
            return self.build_classical_setitem(
                list_wire, idx, elem, elem_ty_arg.ty.to_hugr()
            )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListPopCompiler(CustomCallCompiler):
    """Compiler for the `list.pop` function."""

    def build_classical_pop(
        self,
        list_wire: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.pop` for classical lists."""
        list_wire, result = self.builder.add_op(list_pop(elem_ty), list_wire)
        elem = build_unwrap(self.builder, result, "List index out of bounds")
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def build_linear_pop(
        self,
        list_wire: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.pop` for linear lists."""
        elem_opt_ty = ht.Option(elem_ty)
        list_wire, result = self.builder.add_op(list_pop(elem_opt_ty), list_wire)
        elem_opt = build_unwrap(self.builder, result, "List index out of bounds")
        elem = build_unwrap(
            self.builder, elem_opt, "Linear list element has already been used"
        )
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_pop(list_wire, elem_ty_arg.ty.to_hugr())
        else:
            return self.build_classical_pop(list_wire, elem_ty_arg.ty.to_hugr())

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListPushCompiler(CustomCallCompiler):
    """Compiler for the `list.push` function."""

    def build_classical_push(
        self,
        list_wire: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.push` for classical lists."""
        list_wire = self.builder.add_op(list_push(elem_ty), list_wire, elem)
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def build_linear_push(
        self,
        list_wire: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.push` for linear lists."""
        # Wrap element into an optional
        elem_opt_ty = ht.Option(elem_ty)
        elem_opt = self.builder.add_op(ops.Some(elem_ty), elem)
        list_wire = self.builder.add_op(list_push(elem_opt_ty), list_wire, elem_opt)
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire, elem] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_push(list_wire, elem, elem_ty_arg.ty.to_hugr())
        else:
            return self.build_classical_push(list_wire, elem, elem_ty_arg.ty.to_hugr())

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListLengthCompiler(CustomCallCompiler):
    """Compiler for the `list.__len__` function."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        elem_ty = elem_ty_arg.ty.to_hugr()
        if elem_ty_arg.ty.linear:
            elem_ty = ht.Option(elem_ty)
        list_wire, length = self.builder.add_op(list_length(elem_ty), list_wire)
        length = self.builder.add_op(convert_ifromusize(), length)
        return CallReturnWires(regular_returns=[length], inout_returns=[list_wire])

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


def list_new(
    builder: DfBase[ops.DfParentOp], elem_type: ht.Type, args: list[Wire]
) -> Wire:
    if elem_type.type_bound() == ht.TypeBound.Any:
        return _list_new_linear(builder, elem_type, args)
    else:
        return _list_new_classical(builder, elem_type, args)


def _list_new_classical(
    builder: DfBase[ops.DfParentOp], elem_type: ht.Type, args: list[Wire]
) -> Wire:
    # This may be simplified in the future with a `new` or `with_capacity` list op
    # See https://github.com/CQCL/hugr/issues/1508
    lst = builder.load(ListVal([], elem_ty=elem_type))
    push_op = list_push(elem_type)
    for elem in args:
        lst = builder.add_op(push_op, lst, elem)
    return lst


def _list_new_linear(
    builder: DfBase[ops.DfParentOp], elem_type: ht.Type, args: list[Wire]
) -> Wire:
    elem_opt_ty = ht.Option(elem_type)
    lst = builder.load(ListVal([], elem_ty=elem_opt_ty))
    push_op = list_push(elem_opt_ty)
    for elem in args:
        elem_opt = builder.add_op(ops.Some(elem_type), elem)
        lst = builder.add_op(push_op, lst, elem_opt)
    return lst



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.option import Option
from guppylang.std.quantum import qubit, measure

import guppylang.std.quantum_functional as quantum_functional
from guppylang.std.quantum_functional import cx, t, h, project_z
from guppylang.tys.ty import NoneType



def test_id(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q: qubit @owned) -> qubit:
        return q

    validate(module.compile())


def test_assign(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q: qubit @owned) -> qubit:
        r = q
        s = r
        return s

    validate(module.compile())


def test_linear_return_order(validate):
    # See https://github.com/CQCL-DEV/guppy/issues/35
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q: qubit @owned) -> tuple[qubit, bool]:
        return project_z(q)

    validate(module.compile())


def test_interleave(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.declare(module)
    def f(q1: qubit @owned, q2: qubit @owned) -> tuple[qubit, qubit]: ...

    @guppy.declare(module)
    def g(q1: qubit @owned, q2: qubit @owned) -> tuple[qubit, qubit]: ...

    @guppy(module)
    def test(
        a: qubit @owned, b: qubit @owned, c: qubit @owned, d: qubit @owned
    ) -> tuple[qubit, qubit, qubit, qubit]:
        a, b = f(a, b)
        c, d = f(c, d)
        b, c = g(b, c)
        return a, b, c, d

    validate(module.compile())


def test_linear_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def f(q1: qubit @owned, q2: qubit @owned) -> tuple[qubit, qubit]: ...

    @guppy.declare(module)
    def g(q1: qubit @owned, q2: qubit @owned) -> tuple[qubit, qubit]: ...

    @guppy(module)
    def construct(q1: qubit @owned, q2: qubit @owned) -> MyStruct:
        return MyStruct(q1, q2)

    @guppy(module)
    def test(s: MyStruct @owned, t: MyStruct @owned) -> tuple[qubit, qubit, qubit, qubit]:
        s.q1, s.q2 = f(s.q2, s.q1)
        t.q1, t.q2 = f(t.q1, t.q2)
        s.q2, t.q1 = g(t.q1, s.q2)
        return s.q1, s.q2, t.q1, t.q2

    validate(module.compile())


def test_mixed_classical_linear_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q: qubit
        x: int
        y: float

    @guppy.declare(module)
    def f(q: qubit @owned) -> qubit: ...

    @guppy(module)
    def test1(s: MyStruct @owned) -> tuple[MyStruct, float]:
        a = s.x + s.y
        s.q = f(s.q)
        return s, a * s.x

    @guppy(module)
    def test2(s: MyStruct @owned) -> tuple[MyStruct, int, int, int]:
        t = s
        u = t
        u.q = f(u.q)
        return u, s.x, t.x, u.x

    validate(module.compile())


def test_drop_classical_field(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q: qubit
        x: int

    @guppy.declare(module)
    def f() -> MyStruct: ...

    @guppy(module)
    def test() -> qubit:
        return f().q


def test_if(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(b: bool) -> qubit:
        if b:
            a = qubit()
            q = h(a)
        else:
            q = qubit()
        return q

    validate(module.compile())


def test_if_return(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(b: bool) -> qubit:
        if b:
            q = qubit()
            return h(q)
        else:
            q = qubit()
        return q

    validate(module.compile())


def test_if_struct_rebuild(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q: qubit

    @guppy(module)
    def test1(s: MyStruct @owned, b: bool) -> MyStruct:
        if b:
            s = MyStruct(s.q)
        return s

    @guppy(module)
    def test2(s: MyStruct @owned, b: bool) -> MyStruct:
        if b:
            s.q = s.q
        return s

    validate(module.compile())


def test_struct_reassign(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct1:
        x: "MyStruct2"

    @guppy.struct(module)
    class MyStruct2:
        q: qubit

    @guppy.declare(module)
    def consume(s: MyStruct2 @owned) -> None: ...

    @guppy(module)
    def test(s: MyStruct2 @owned, b: bool) -> MyStruct2:
        consume(s)
        if b:
            s = MyStruct2(qubit())
            return s
        else:
            s = MyStruct1(MyStruct2(qubit()))
        return s.x


def test_struct_reassign2(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def use(q: qubit @owned) -> None: ...

    @guppy(module)
    def test(s: MyStruct @owned, b: bool) -> MyStruct:
        use(s.q1)
        if b:
            s.q1 = qubit()
        else:
            s.q1 = qubit()
        s.q2 = qubit()
        return s


def test_measure(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit, measure)

    @guppy(module)
    def test(q: qubit @owned, x: int) -> int:
        b = measure(q)
        return x

    validate(module.compile())


def test_return_call(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.declare(module)
    def op(q: qubit @owned) -> qubit: ...

    @guppy(module)
    def test(q: qubit @owned) -> qubit:
        return op(q)

    validate(module.compile())


def test_while(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q: qubit @owned, i: int) -> qubit:
        while i > 0:
            i -= 1
            q = h(q)
        return q

    validate(module.compile())


def test_while_break(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q: qubit @owned, i: int) -> qubit:
        while i > 0:
            i -= 1
            q = h(q)
            if i < 5:
                break
        return q

    validate(module.compile())


def test_while_continue(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q: qubit @owned, i: int) -> qubit:
        while i > 0:
            i -= 1
            if i % 3 == 0:
                continue
            q = h(q)
        return q

    validate(module.compile())


def test_while_reset(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def foo(i: bool) -> bool:
        b = False
        while True:
            q = qubit()
            b ^= measure(q)
            if i == 0:
                break
            i -= 1
        return b


def test_while_move_back(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.struct(module)
    class MyStruct:
        q: qubit

    @guppy.declare(module)
    def use(q: qubit @owned) -> None: ...

    @guppy(module)
    def test(s: MyStruct @owned) -> MyStruct:
        use(s.q)
        while True:
            s.q = qubit()
            return s

    validate(module.compile())


def test_for(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(qs: list[tuple[qubit, qubit]] @owned) -> list[qubit]:
        rs: list[qubit] = []
        for q1, q2 in qs:
            q1, q2 = cx(q1, q2)
            rs.append(q1)
            rs.append(q2)
        return rs

    validate(module.compile())


def test_for_measure(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit, measure)

    @guppy(module)
    def test(qs: list[qubit] @owned) -> bool:
        parity = False
        for q in qs:
            parity |= measure(q)
        return parity

    validate(module.compile())


def test_for_continue(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit, measure)

    @guppy(module)
    def test(qs: list[qubit] @owned) -> int:
        x = 0
        for q in qs:
            if measure(q):
                continue
            x += 1
        return x

    validate(module.compile())


def test_for_nonlinear_break(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy.type(NoneType().to_hugr(), module=module)
    class MyIter:
        """An iterator that yields linear values but is not linear itself."""

        @guppy.declare(module)
        def __next__(self: "MyIter") -> Option[tuple[qubit, "MyIter"]]: ...


    @guppy.type(NoneType().to_hugr(), module=module)
    class MyType:
        """Type that produces the iterator above."""

        @guppy.declare(module)
        def __iter__(self: "MyType") -> MyIter: ...

    @guppy.declare(module)
    def measure(q: qubit @owned) -> bool: ...

    @guppy(module)
    def test(mt: MyType, xs: list[int]) -> None:
        # We can break, since `mt` itself is not linear
        for q in mt:
            if measure(q):
                break

    validate(module.compile())


def test_rus(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit, measure)

    @guppy(module)
    def repeat_until_success(q: qubit @owned) -> qubit:
        while True:
            aux, q = cx(t(h(qubit())), q)
            aux, q = cx(h(aux), q)
            if measure(h(t(aux))):
                break
        return q

    validate(module.compile())

def test_list_iter_arg(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def owned_arg(qs: list[qubit] @owned) -> list[qubit]:
        qs = [h(q) for q in qs]
        return qs

    validate(module.compile())

def test_list_iter(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def owned_arg() -> list[qubit]:
        qs = [qubit() for _ in [0,1,2]]
        qs = [h(q) for q in qs]
        return qs

    validate(module.compile())


def test_non_terminating(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test() -> None:
        q = qubit()
        while True:
            q = h(q)

    validate(module.compile())



"""Guppy standard module for dyadic rational angles."""

# mypy: disable-error-code="empty-body, misc, override, operator"

import math
from typing import no_type_check

from hugr import val as hv
from hugr.std.float import FloatVal

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import py

angles = GuppyModule("angles")


@guppy.struct(angles)
class angle:
    """Not an angle in the truest sense but a rotation by a number of half-turns
    (does not wrap or identify with itself modulo any number of complete turns).
    """

    halfturns: float

    @guppy(angles)
    @no_type_check
    def __add__(self: "angle", other: "angle") -> "angle":
        return angle(self.halfturns + other.halfturns)

    @guppy(angles)
    @no_type_check
    def __sub__(self: "angle", other: "angle") -> "angle":
        return angle(self.halfturns - other.halfturns)

    @guppy(angles)
    @no_type_check
    def __mul__(self: "angle", other: float) -> "angle":
        return angle(self.halfturns * other)

    @guppy(angles)
    @no_type_check
    def __rmul__(self: "angle", other: float) -> "angle":
        return angle(self.halfturns * other)

    @guppy(angles)
    @no_type_check
    def __truediv__(self: "angle", other: float) -> "angle":
        return angle(self.halfturns / other)

    @guppy(angles)
    @no_type_check
    def __rtruediv__(self: "angle", other: float) -> "angle":
        return angle(other / self.halfturns)

    @guppy(angles)
    @no_type_check
    def __neg__(self: "angle") -> "angle":
        return angle(-self.halfturns)

    @guppy(angles)
    @no_type_check
    def __float__(self: "angle") -> float:
        return self.halfturns * py(math.pi)

    @guppy(angles)
    @no_type_check
    def __eq__(self: "angle", other: "angle") -> bool:
        return self.halfturns == other.halfturns


pi = guppy.constant("pi", ty="angle", value=hv.Tuple(FloatVal(1.0)), module=angles)



import pytest

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


def test_func_redefinition(validate):
    module = GuppyModule("test")

    @guppy(module)
    def test() -> bool:
        return 5  # Type error on purpose

    @guppy(module)
    def test() -> bool:  # noqa: F811
        return False

    validate(module.compile())


def test_method_redefinition(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class Test:
        x: int

        @guppy(module)
        def foo(self: "Test") -> int:
            return 1.0  # Type error on purpose

        @guppy(module)
        def foo(self: "Test") -> int:
            return 1  # Type error on purpose

    validate(module.compile())


def test_redefine_after_error(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class Foo:
        x: int

    @guppy(module)
    def foo() -> int:
        return y

    try:
        module.compile()
    except:
        pass

    @guppy.struct(module)
    class Foo:
        x: int

    @guppy(module)
    def foo(f: Foo) -> int:
        return f.x

    validate(module.compile())


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/456")
def test_struct_redefinition(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class Test:
        x: "blah"  # Non-existing type

    @guppy.struct(module)
    class Test:
        y: int

    @guppy(module)
    def main(x: int) -> Test:
        return Test(x)

    validate(module.compile())


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/456")
def test_struct_method_redefinition(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class Test:
        x: int

        @guppy(module)
        def foo(self: "Test") -> int:
            return 1.0  # Type error on purpose

    @guppy.struct(module)
    class Test:
        y: int

        @guppy(module)
        def bar(self: "Test") -> int:
            return self.y

    @guppy(module)
    def main(x: int) -> int:
        return Test(x).bar()

    validate(module.compile())




import ast
import copy
import itertools
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field, replace
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    NamedTuple,
    TypeAlias,
    TypeVar,
)

from typing_extensions import assert_never

from guppylang.ast_util import AstNode, name_nodes_in_ast
from guppylang.cfg.bb import VId
from guppylang.definition.common import DefId, Definition
from guppylang.definition.ty import TypeDef
from guppylang.definition.value import CallableDef
from guppylang.tys.builtin import (
    array_type_def,
    bool_type_def,
    callable_type_def,
    float_type_def,
    int_type_def,
    list_type_def,
    nat_type_def,
    none_type_def,
    option_type_def,
    sized_iter_type_def,
    string_type_def,
    tuple_type_def,
)
from guppylang.tys.param import Parameter
from guppylang.tys.ty import (
    BoundTypeVar,
    ExistentialTypeVar,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    OpaqueType,
    StructType,
    SumType,
    TupleType,
    Type,
)

if TYPE_CHECKING:
    from guppylang.definition.struct import StructField


#: A "place" is a description for a storage location of a local value that users
#: can refer to in their program.
#:
#: Roughly, these are values that can be lowered to a static wire within the Hugr
#: representation. The most basic example of a place is a single local variable. Beyond
#: that, we also treat some projections of local variables (e.g. nested struct field
#: accesses) as places.
#:
#: All places are equipped with a unique id, a type and an optional definition AST
#: location. During linearity checking, they are tracked separately.
Place: TypeAlias = "Variable | FieldAccess | SubscriptAccess"

#: Unique identifier for a `Place`.
PlaceId: TypeAlias = "Variable.Id | FieldAccess.Id | SubscriptAccess.Id"


@dataclass(frozen=True)
class Variable:
    """A place identifying a local variable."""

    name: str
    ty: Type
    defined_at: AstNode | None
    flags: InputFlags = InputFlags.NoFlags

    @dataclass(frozen=True)
    class Id:
        """Identifier for variable places."""

        name: str

    @cached_property
    def id(self) -> "Variable.Id":
        """The unique `PlaceId` identifier for this place."""
        return Variable.Id(self.name)

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self

    @property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Variable `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return self.name

    def replace_defined_at(self, node: AstNode | None) -> "Variable":
        """Returns a new `Variable` instance with an updated definition location."""
        return replace(self, defined_at=node)


@dataclass(frozen=True)
class FieldAccess:
    """A place identifying a field access on a local struct."""

    parent: Place
    field: "StructField"
    exact_defined_at: AstNode | None

    @dataclass(frozen=True)
    class Id:
        """Identifier for field places."""

        parent: PlaceId
        field: str

    def __post_init__(self) -> None:
        # Check that the field access is consistent
        assert self.struct_ty.field_dict[self.field.name] == self.field

    @cached_property
    def id(self) -> "FieldAccess.Id":
        """The unique `PlaceId` identifier for this place."""
        return FieldAccess.Id(self.parent.id, self.field.name)

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self.parent.root

    @property
    def ty(self) -> Type:
        """The type of this place."""
        return self.field.ty

    @cached_property
    def struct_ty(self) -> StructType:
        """The type of the struct whose field is accessed."""
        assert isinstance(self.parent.ty, StructType)
        return self.parent.ty

    @cached_property
    def defined_at(self) -> AstNode | None:
        """Optional location where this place was last assigned to."""
        return self.exact_defined_at or self.parent.defined_at

    @cached_property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Field `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return f"{self.parent}.{self.field.name}"

    def replace_defined_at(self, node: AstNode | None) -> "FieldAccess":
        """Returns a new `FieldAccess` instance with an updated definition location."""
        return replace(self, exact_defined_at=node)


class SetitemCall(NamedTuple):
    """
    Represents a `__setitem__` call for assigning values to array elements.

    Holds the expression corresponding to the `__setitem__` call and the variable
    holding the value to be written to the subscript.
    """

    #: Expression corresponding to the `__setitem__` call.
    call: ast.expr

    #: Variable holding the value that should be written to the subscript.
    #: This variable *must be* assigned before compiling the call!
    value_var: Variable


@dataclass(frozen=True)
class SubscriptAccess:
    """A place identifying a subscript `place[item]` access."""

    parent: Place
    item: Variable
    ty: Type
    item_expr: ast.expr
    getitem_call: ast.expr | None = None
    setitem_call: SetitemCall | None = None

    @dataclass(frozen=True)
    class Id:
        """Identifier for subscript places."""

        parent: PlaceId
        item: Variable.Id

    @cached_property
    def id(self) -> "SubscriptAccess.Id":
        """The unique `PlaceId` identifier for this place."""
        return SubscriptAccess.Id(self.parent.id, self.item.id)

    @cached_property
    def defined_at(self) -> AstNode | None:
        """Optional location where this place was last assigned to."""
        return self.parent.defined_at

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self.parent.root

    @property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Subscript `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return f"{self.parent}[...]"


def contains_subscript(place: Place) -> SubscriptAccess | None:
    """Checks if a place contains a subscript access and returns the rightmost one."""
    while not isinstance(place, Variable):
        if isinstance(place, SubscriptAccess):
            return place
        place = place.parent
    return None


PyScope = dict[str, Any]


@dataclass(frozen=True)
class Globals:
    """Collection of definitions that are available on module-level.

    Stores a mapping from global ids to their definition together with a mapping of
    user names to definition id and instance implementation id.
    """

    defs: dict[DefId, Definition]

    names: dict[str, DefId]
    impls: dict[DefId, dict[str, DefId]]
    python_scope: PyScope = field(repr=False)

    @staticmethod
    def default() -> "Globals":
        """Generates a `Globals` instance that is populated with all core types"""
        builtins: list[Definition] = [
            callable_type_def,
            tuple_type_def,
            none_type_def,
            bool_type_def,
            nat_type_def,
            int_type_def,
            float_type_def,
            string_type_def,
            list_type_def,
            array_type_def,
            sized_iter_type_def,
            option_type_def,
        ]
        defs = {defn.id: defn for defn in builtins}
        names = {defn.name: defn.id for defn in builtins}
        return Globals(defs, names, {}, {})

    def get_instance_func(self, ty: Type | TypeDef, name: str) -> CallableDef | None:
        """Looks up an instance function with a given name for a type.

        Returns `None` if the name doesn't exist or isn't a function.
        """
        type_defn: TypeDef
        match ty:
            case TypeDef() as type_defn:
                pass
            case BoundTypeVar() | ExistentialTypeVar() | SumType():
                return None
            case NumericType(kind):
                match kind:
                    case NumericType.Kind.Nat:
                        type_defn = nat_type_def
                    case NumericType.Kind.Int:
                        type_defn = int_type_def
                    case NumericType.Kind.Float:
                        type_defn = float_type_def
                    case kind:
                        return assert_never(kind)
            case FunctionType():
                type_defn = callable_type_def
            case OpaqueType() as ty:
                type_defn = ty.defn
            case StructType() as ty:
                type_defn = ty.defn
            case TupleType():
                type_defn = tuple_type_def
            case NoneType():
                type_defn = none_type_def
            case _:
                return assert_never(ty)

        if type_defn.id in self.impls and name in self.impls[type_defn.id]:
            def_id = self.impls[type_defn.id][name]
            defn = self.defs[def_id]
            if isinstance(defn, CallableDef):
                return defn
        return None

    def with_python_scope(self, python_scope: PyScope) -> "Globals":
        return Globals(
            self.defs, self.names, self.impls, self.python_scope | python_scope
        )

    def __or__(self, other: "Globals") -> "Globals":
        impls = {
            def_id: self.impls.get(def_id, {}) | other.impls.get(def_id, {})
            for def_id in self.impls.keys() | other.impls.keys()
        }
        return Globals(
            {**self.defs, **other.defs},  # Can't use `|` since it's a Mapping
            self.names | other.names,
            impls,
            self.python_scope | other.python_scope,
        )

    def __contains__(self, item: DefId | str) -> bool:
        match item:
            case DefId() as def_id:
                return def_id in self.defs
            case str(name):
                return name in self.names
            case x:
                return assert_never(x)

    def __getitem__(self, item: DefId | str) -> Definition:
        match item:
            case DefId() as def_id:
                return self.defs[def_id]
            case str(name):
                return self.defs[self.names[name]]
            case x:
                return assert_never(x)


V = TypeVar("V")


@dataclass
class Locals(Generic[VId, V]):
    """Scoped mapping from program variable ids to the corresponding program variable.

    Depending on which checking phase we are in (type checking or linearity checking),
    we use this either as a mapping from strings to `Variable`s or as a mapping from
    `PlaceId`s to `Place`s.
    """

    vars: dict[VId, V]
    parent_scope: "Locals[VId, V] | None" = None

    def __getitem__(self, item: VId) -> V:
        if item not in self.vars and self.parent_scope:
            return self.parent_scope[item]

        return self.vars[item]

    def __setitem__(self, key: VId, value: V) -> None:
        self.vars[key] = value

    def __iter__(self) -> Iterator[VId]:
        parent_iter = iter(self.parent_scope) if self.parent_scope else iter(())
        return itertools.chain(iter(self.vars), parent_iter)

    def __contains__(self, item: VId) -> bool:
        return (item in self.vars) or (
            self.parent_scope is not None and item in self.parent_scope
        )

    def __copy__(self) -> "Locals[VId, V]":
        # Make a copy of the var map so that mutating the copy doesn't
        # mutate our variable mapping
        return Locals(self.vars.copy(), copy.copy(self.parent_scope))

    def keys(self) -> set[VId]:
        parent_keys = self.parent_scope.keys() if self.parent_scope else set()
        return parent_keys | self.vars.keys()

    def values(self) -> Iterable[V]:
        parent_values = (
            iter(self.parent_scope.values()) if self.parent_scope else iter(())
        )
        return itertools.chain(self.vars.values(), parent_values)

    def items(self) -> Iterable[tuple[VId, V]]:
        parent_items = (
            iter(self.parent_scope.items()) if self.parent_scope else iter(())
        )
        return itertools.chain(self.vars.items(), parent_items)


class Context(NamedTuple):
    """The type checking context."""

    globals: Globals
    locals: Locals[str, Variable]
    generic_params: dict[str, Parameter]


class DummyEvalDict(PyScope):
    """A custom dict that can be passed to `eval` to give better error messages.
    This class is used to implement the `py(...)` expression. If the user tries to
    access a Guppy variable in the Python context, we give an informative error message.
    """

    ctx: Context
    node: ast.expr

    @dataclass
    class GuppyVarUsedError(BaseException):
        """Error that is raised when the user tries to access a Guppy variable."""

        var: str
        node: ast.Name | None

    def __init__(self, ctx: Context, node: ast.expr):
        super().__init__(**ctx.globals.python_scope)
        self.ctx = ctx
        self.node = node

    def _check_item(self, key: str) -> None:
        # Catch the user trying to access Guppy variables
        if key in self.ctx.locals:
            # Find the name node in the AST where the usage occurs
            n = next((n for n in name_nodes_in_ast(self.node) if n.id == key), None)
            raise self.GuppyVarUsedError(key, n)

    def __getitem__(self, key: str) -> Any:
        self._check_item(key)
        return super().__getitem__(key)

    def __delitem__(self, key: str) -> None:
        self._check_item(key)
        super().__delitem__(key)

    def __contains__(self, key: object) -> bool:
        if isinstance(key, str) and key in self.ctx.locals:
            return True
        return super().__contains__(key)



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)


@guppy.struct(module)
class Struct:
    q: qubit

    @guppy(module)
    def foo(self: "Struct @owned") -> "Struct":
        return self


@guppy(module)
def foo(s: Struct @owned) -> Struct:
    f = s.foo
    return f()


module.compile()



from guppylang.decorator import guppy
from guppylang.std.angles import angle, pi, angles
from guppylang.std.builtins import nat
from guppylang.module import GuppyModule
from tests.util import compile_guppy


def test_negative(validate, run_int_fn):
    module = GuppyModule("test_negative")

    @guppy(module)
    def negative() -> int:
        return 2 - 42

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=-40, fn_name="negative")


def test_arith_basic(validate, run_int_fn):
    module = GuppyModule("test_arith")

    @guppy(module)
    def add(x: int, y: int) -> int:
        return x + y

    @guppy(module)
    def main() -> int:
        return add(40, 2)

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, 42)


def test_constant(validate):
    @compile_guppy
    def const() -> float:
        return 42.0

    validate(const)


def test_nat_literal(validate):
    @compile_guppy
    def const() -> nat:
        return 42

    validate(const)


def test_aug_assign(validate, run_int_fn):
    module = GuppyModule("test_aug_assign")

    @guppy(module)
    def add(x: int) -> int:
        x += 1
        return x

    @guppy(module)
    def main() -> int:
        return add(5)

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, 6)


def test_nat(validate):
    @compile_guppy
    def foo(
        a: nat, b: nat, c: bool, d: int, e: float
    ) -> tuple[nat, bool, int, float, float]:
        b, c, d, e = nat(b), nat(c), nat(d), nat(e)
        x = a + b * c // d - e
        y = e / b
        return x, bool(x), int(x), float(x), y

    validate(foo)


def test_nat2(validate):
    @compile_guppy
    def foo(
        a: nat, b: nat, c: bool, d: int, e: float
    ) -> tuple[nat, bool, int, float, float]:
        b, c, d, e = nat(b), nat(c), nat(d), nat(e)
        x = a + b * c // d - e
        y = e / b
        return x, bool(x), int(x), float(x), y

    validate(foo)


def test_float_coercion(validate):
    @compile_guppy
    def coerce(x: int, y: float) -> float:
        return x * y

    validate(coerce)


def test_arith_big(validate):
    @compile_guppy
    def arith(x: int, y: float, z: int) -> bool:
        a = x // y + 3 * z
        b = -8 >= a > 5 or (x * y == 0 and a % 3 < x)
        return b

    validate(arith)


def test_angle_arith(validate):
    module = GuppyModule("test")
    module.load(angle)

    @guppy(module)
    def main(a1: angle, a2: angle) -> bool:
        a3 = -a1 + a2 * -3
        a3 -= a1
        a3 += 2 * a1
        return a3 / 3 == -a2

    validate(module.compile())


def test_angle_arith_float(validate):
    module = GuppyModule("test")
    module.load(angle)

    @guppy(module)
    def main(a1: angle, a2: angle) -> bool:
        a3 = -a1 + a2 * -3.5
        a3 -= a1
        a3 += 2.2 * a1
        return a3 / 3.9 == -a2

    validate(module.compile())


def test_implicit_coercion(validate):
    @compile_guppy
    def coerce(x: nat) -> float:
        y: int = x
        z: float = y
        a: float = 1
        return z + a

    validate(coerce)


def test_angle_float_coercion(validate):
    module = GuppyModule("test")
    module.load(angle)

    @guppy(module)
    def main(f: float) -> tuple[angle, float]:
        a = angle(f)
        return a, float(a)

    validate(module.compile())


def test_angle_pi(validate):
    module = GuppyModule("test")
    module.load(angle, pi)

    @guppy(module)
    def main() -> angle:
        a = 2 * pi
        a += -pi / 3
        a += 3 * pi / 2
        return a

    validate(module.compile())


def test_shortcircuit_assign1(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> bool:
        if (z := x) and y > 0:
            return z
        return not z

    validate(foo)


def test_shortcircuit_assign2(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> bool:
        if y > 0 and (z := x):
            return z
        return False

    validate(foo)


def test_shortcircuit_assign3(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> bool:
        if (z := x) or y > 0:
            return z
        return z

    validate(foo)


def test_shortcircuit_assign4(validate):
    @compile_guppy
    def foo(x: bool, y: int) -> bool:
        if y > 0 or (z := x):
            return False
        return z

    validate(foo)


def test_supported_ops(validate, run_int_fn):
    module = GuppyModule("supported_ops")

    @guppy(module)
    def double_add(x: int) -> int:
        return x + x

    @guppy(module)
    def double_mul(x: int) -> int:
        return x * 2

    @guppy(module)
    def quad(x: int) -> int:
        y = double_add(x)
        z = double_mul(x)
        return y + z

    @guppy(module)
    def run_quad() -> int:
        return quad(42)

    @guppy(module)
    def neg(x: int) -> int:
        return -x

    @guppy(module)
    def run_neg() -> int:
        return neg(42)

    @guppy(module)
    def div(x: int, y: int) -> int:
        return x // y

    @guppy(module)
    def run_div() -> int:
        return div(-42, 21)

    @guppy(module)
    def run_rem() -> int:
        return 11 % 3

    hugr = module.compile()
    validate(hugr)
    run_int_fn(hugr, expected=168, fn_name="run_quad")
    run_int_fn(hugr, expected=-42, fn_name="run_neg")
    run_int_fn(hugr, expected=-2, fn_name="run_div")
    run_int_fn(hugr, expected=2, fn_name="run_rem")


def test_angle_exec(validate, run_float_fn_approx):
    module = GuppyModule("test_angle_exec")
    module.load_all(angles)

    @guppy(module)
    def main() -> float:
        a1 = pi
        a2 = pi * 2
        a3 = -a1 + a2 * -3
        a3 -= a1
        a3 += 2 * a1
        return float(a3)

    hugr = module.compile()
    validate(hugr)
    import math

    run_float_fn_approx(hugr, expected=-6 * math.pi)


def test_xor(validate, run_int_fn):
    module = GuppyModule("test_xor")

    @guppy(module)
    def main() -> int:
        return int(True ^ False ^ False)

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, 1)

    module = GuppyModule("test_xor")

    @guppy(module)
    def main() -> int:
        return int(True ^ False ^ False ^ True)

    compiled = module.compile()
    run_int_fn(compiled, 0)


def test_pow(validate, run_int_fn) -> None:
    module = GuppyModule("test_pow")

    @guppy(module)
    def main() -> int:
        return -(3**3) + 4**2

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, -11)


def test_float_to_int(validate, run_int_fn) -> None:
    module = GuppyModule("test_cast")

    @guppy(module)
    def main() -> int:
        return int(-2.75)

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, -2)


def test_float_to_nat(validate, run_int_fn) -> None:
    module = GuppyModule("test_cast")

    @guppy(module)
    def main() -> nat:
        return nat(2.75)

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, 2)
